#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import sys
import unittest

from random import shuffle
from xotl.tools.future.collections import defaultdict


class TestCollections(unittest.TestCase):
    def test_defaultdict(self):
        d = defaultdict(lambda key, _: "a")
        self.assertEqual("a", d["abc"])
        d["abc"] = 1
        self.assertEqual(1, d["abc"])

    def test_defaultdict_clone(self):
        d = defaultdict(lambda key, d: d["a"], {"a": "default"})
        self.assertEqual("default", d["abc"])

        d = defaultdict(lambda key, d: d[key])
        with self.assertRaises(KeyError):
            d["abc"]


def test_stacked_dict_with_newpop():
    """Test that stacked.pop has the same semantics has dict.pop."""
    from xotl.tools.future.collections import StackedDict

    sd = StackedDict(a="level-0", b=1)
    assert sd.pop("a") == "level-0"
    assert sd.pop("non", sd) is sd
    try:
        sd.pop("non")
    except KeyError:
        pass
    else:
        assert False, "Should have raised a KeyError"


def test_stacked_dict():
    from xotl.tools.future.collections import StackedDict

    sd = StackedDict(a="level-0")
    assert sd.peek() == dict(a="level-0")
    sd.push_level(a=1, b=2, c=10)
    assert sd.level == 1
    assert sd.peek() == dict(a=1, b=2, c=10)
    sd.push_level(b=4, c=5)
    assert sd.peek() == dict(b=4, c=5)
    assert sd.level == 2
    assert sd["b"] == 4
    assert sd["a"] == 1
    assert sd["c"] == 5
    assert len(sd) == 3
    del sd["c"]
    try:
        del sd["c"]
        assert False, "Should have raise KeyError"
    except KeyError:
        pass
    except:
        assert False, "Should have raise KeyError"
    assert sd.pop_level() == {"b": 4}
    assert sd["b"] == 2
    assert sd["a"] == 1
    assert len(sd) == 3
    sd.pop_level()
    assert sd["a"] == "level-0"
    try:
        sd.pop_level()
        assert False, "Level 0 cannot be poped. " "It should have raised a TypeError"
    except TypeError:
        pass
    except:
        assert False, "Level 0 cannot be poped. " "It should have raised a TypeError"


# Backported from Python 3.3.0 standard library
from xotl.tools.future.collections import ChainMap, Counter
from xotl.tools.future.collections import OrderedDict, RankedDict
from xotl.tools.future.collections import Mapping, MutableMapping
import copy
import pickle
from random import randrange


def _items(d):
    "For some reason in new PyPy 5.0.1 for Py 2.7.10, set order is not nice."
    from xotl.tools.versions import python_version

    res = d.items()
    if python_version.pypy and isinstance(res, list):
        res.sort()
    return res


class TestChainMap(unittest.TestCase):
    def test_basics(self):
        c = ChainMap()
        c["a"] = 1
        c["b"] = 2
        d = c.new_child()
        d["b"] = 20
        d["c"] = 30
        # check internal state
        self.assertEqual(d.maps, [{"b": 20, "c": 30}, {"a": 1, "b": 2}])
        # check items/iter/getitem
        self.assertEqual(_items(d), _items(dict(a=1, b=20, c=30)))
        # check len
        self.assertEqual(len(d), 3)
        # check contains
        for key in "abc":
            self.assertIn(key, d)
        # check get
        for k, v in dict(a=1, b=20, c=30, z=100).items():
            self.assertEqual(d.get(k, 100), v)

        # unmask a value
        del d["b"]
        # check internal state
        self.assertEqual(d.maps, [{"c": 30}, {"a": 1, "b": 2}])
        # check items/iter/getitem
        self.assertEqual(_items(d), _items(dict(a=1, b=2, c=30)))
        # check len
        self.assertEqual(len(d), 3)
        # check contains
        for key in "abc":
            self.assertIn(key, d)
        # check get
        for k, v in dict(a=1, b=2, c=30, z=100).items():
            self.assertEqual(d.get(k, 100), v)
        # check repr
        self.assertIn(
            repr(d),
            [
                type(d).__name__ + "({'c': 30}, {'a': 1, 'b': 2})",
                type(d).__name__ + "({'c': 30}, {'b': 2, 'a': 1})",
            ],
        )

        # check shallow copies
        for e in d.copy(), copy.copy(d):
            self.assertEqual(d, e)
            self.assertEqual(d.maps, e.maps)
            self.assertIsNot(d, e)
            self.assertIsNot(d.maps[0], e.maps[0])
            for m1, m2 in zip(d.maps[1:], e.maps[1:]):
                self.assertIs(m1, m2)

        # check deep copies
        for e in [pickle.loads(pickle.dumps(d)), copy.deepcopy(d), eval(repr(d))]:
            self.assertEqual(d, e)
            self.assertEqual(d.maps, e.maps)
            self.assertIsNot(d, e)
            for m1, m2 in zip(d.maps, e.maps):
                self.assertIsNot(m1, m2, e)

        f = d.new_child()
        f["b"] = 5
        self.assertEqual(f.maps, [{"b": 5}, {"c": 30}, {"a": 1, "b": 2}])
        # check parents
        self.assertEqual(f.parents.maps, [{"c": 30}, {"a": 1, "b": 2}])
        # find first in chain
        self.assertEqual(f["b"], 5)
        # look beyond maps[0]
        self.assertEqual(f.parents["b"], 2)

    def test_contructor(self):
        # no-args --> one new dict
        self.assertEqual(ChainMap().maps, [{}])
        # 1 arg --> list
        self.assertEqual(ChainMap({1: 2}).maps, [{1: 2}])

    def test_bool(self):
        self.assertFalse(ChainMap())
        self.assertFalse(ChainMap({}, {}))
        self.assertTrue(ChainMap({1: 2}, {}))
        self.assertTrue(ChainMap({}, {1: 2}))

    def test_missing(self):
        class DefaultChainMap(ChainMap):
            def __missing__(self, key):
                return 999

        d = DefaultChainMap(dict(a=1, b=2), dict(b=20, c=30))
        for k, v in dict(a=1, b=2, c=30, d=999).items():
            # check __getitem__ w/missing
            self.assertEqual(d[k], v)
        for k, v in dict(a=1, b=2, c=30, d=77).items():
            # check get() w/ missing
            self.assertEqual(d.get(k, 77), v)
        for k, v in dict(a=True, b=True, c=True, d=False).items():
            # check __contains__ w/missing
            self.assertEqual(k in d, v)
        self.assertEqual(d.pop("a", 1001), 1, d)
        # check pop() w/missing
        self.assertEqual(d.pop("a", 1002), 1002)
        # check popitem() w/missing
        self.assertEqual(d.popitem(), ("b", 2))
        with self.assertRaises(KeyError):
            d.popitem()

    def test_dict_coercion(self):
        d = ChainMap(dict(a=1, b=2), dict(b=20, c=30))
        self.assertEqual(dict(d), dict(a=1, b=2, c=30))
        self.assertEqual(dict(d.items()), dict(a=1, b=2, c=30))

    def test_new_child(self):
        "Tests for changes for issue #16613."
        c = ChainMap()
        c["a"] = 1
        c["b"] = 2
        m = {"b": 20, "c": 30}
        d = c.new_child(m)
        # check internal state
        self.assertEqual(d.maps, [{"b": 20, "c": 30}, {"a": 1, "b": 2}])
        self.assertIs(m, d.maps[0])

        # Use a different map than a dict
        class lowerdict(dict):
            def __getitem__(self, key):
                if isinstance(key, str):
                    key = key.lower()
                return dict.__getitem__(self, key)

            def __contains__(self, key):
                if isinstance(key, str):
                    key = key.lower()
                return dict.__contains__(self, key)

        c = ChainMap()
        c["a"] = 1
        c["b"] = 2
        m = lowerdict(b=20, c=30)
        d = c.new_child(m)
        self.assertIs(m, d.maps[0])
        # check contains
        for key in "abc":
            self.assertIn(key, d)
        # check get
        for k, v in dict(a=1, B=20, C=30, z=100).items():
            self.assertEqual(d.get(k, 100), v)


class TestCounter(unittest.TestCase):
    def test_basics(self):
        c = Counter("abcaba")
        self.assertEqual(c, Counter({"a": 3, "b": 2, "c": 1}))
        self.assertEqual(c, Counter(a=3, b=2, c=1))
        self.assertIsInstance(c, dict)
        self.assertIsInstance(c, Mapping)
        self.assertTrue(issubclass(Counter, dict))
        self.assertTrue(issubclass(Counter, Mapping))
        self.assertEqual(len(c), 3)
        self.assertEqual(sum(c.values()), 6)
        self.assertEqual(sorted(c.values()), [1, 2, 3])
        self.assertEqual(sorted(c.keys()), ["a", "b", "c"])
        self.assertEqual(sorted(c), ["a", "b", "c"])
        self.assertEqual(sorted(c.items()), [("a", 3), ("b", 2), ("c", 1)])
        self.assertEqual(c["b"], 2)
        self.assertEqual(c["z"], 0)
        self.assertEqual(c.__contains__("c"), True)
        self.assertEqual(c.__contains__("z"), False)
        self.assertEqual(c.get("b", 10), 2)
        self.assertEqual(c.get("z", 10), 10)
        self.assertEqual(c, dict(a=3, b=2, c=1))
        self.assertEqual(repr(c), "Counter({'a': 3, 'b': 2, 'c': 1})")
        self.assertEqual(c.most_common(), [("a", 3), ("b", 2), ("c", 1)])
        for i in range(5):
            self.assertEqual(c.most_common(i), [("a", 3), ("b", 2), ("c", 1)][:i])
        self.assertEqual("".join(sorted(c.elements())), "aaabbc")
        c["a"] += 1  # increment an existing value
        c["b"] -= 2  # sub existing value to zero
        del c["c"]  # remove an entry
        del c["c"]  # make sure that del doesn't raise KeyError
        c["d"] -= 2  # sub from a missing value
        c["e"] = -5  # directly assign a missing value
        c["f"] += 4  # add to a missing value
        self.assertEqual(c, dict(a=4, b=0, d=-2, e=-5, f=4))
        self.assertEqual("".join(sorted(c.elements())), "aaaaffff")
        self.assertEqual(c.pop("f"), 4)
        self.assertNotIn("f", c)
        for i in range(3):
            elem, cnt = c.popitem()
            self.assertNotIn(elem, c)
        c.clear()
        self.assertEqual(c, {})
        self.assertEqual(repr(c), "Counter()")
        self.assertRaises(NotImplementedError, Counter.fromkeys, "abc")
        self.assertRaises(TypeError, hash, c)
        c.update(dict(a=5, b=3))
        c.update(c=1)
        c.update(Counter("a" * 50 + "b" * 30))
        c.update()  # test case with no args
        c.__init__("a" * 500 + "b" * 300)
        c.__init__("cdc")
        c.__init__()
        self.assertEqual(c, dict(a=555, b=333, c=3, d=1))
        self.assertEqual(c.setdefault("d", 5), 1)
        self.assertEqual(c["d"], 1)
        self.assertEqual(c.setdefault("e", 5), 5)
        self.assertEqual(c["e"], 5)

    def test_copying(self):
        # Check that counters are copyable, deepcopyable, picklable, and
        # have a repr/eval round-trip
        words = Counter("which witch had which witches wrist watch".split())
        update_test = Counter()
        update_test.update(words)
        for i, dup in enumerate(
            [
                words.copy(),
                copy.copy(words),
                copy.deepcopy(words),
                pickle.loads(pickle.dumps(words, 0)),
                pickle.loads(pickle.dumps(words, 1)),
                pickle.loads(pickle.dumps(words, 2)),
                pickle.loads(pickle.dumps(words, -1)),
                eval(repr(words)),
                update_test,
                Counter(words),
            ]
        ):
            # TODO: Not used ``msg = (i, dup, words)``
            self.assertTrue(dup is not words)
            self.assertEqual(dup, words)
            self.assertEqual(len(dup), len(words))
            self.assertEqual(type(dup), type(words))

    def test_copy_subclass(self):
        class MyCounter(Counter):
            pass

        c = MyCounter("slartibartfast")
        d = c.copy()
        self.assertEqual(d, c)
        self.assertEqual(len(d), len(c))
        self.assertEqual(type(d), type(c))

    def test_conversions(self):
        # Convert to: set, list, dict
        s = "she sells sea shells by the sea shore"
        self.assertEqual(sorted(Counter(s).elements()), sorted(s))
        self.assertEqual(sorted(Counter(s)), sorted(set(s)))
        self.assertEqual(dict(Counter(s)), dict(Counter(s).items()))
        self.assertEqual(set(Counter(s)), set(s))

    def test_invariant_for_the_in_operator(self):
        c = Counter(a=10, b=-2, c=0)
        for elem in c:
            self.assertTrue(elem in c)
            self.assertIn(elem, c)

    def test_multiset_operations(self):
        # Verify that adding a zero counter will strip zeros and negatives
        c = Counter(a=10, b=-2, c=0) + Counter()
        self.assertEqual(dict(c), dict(a=10))

        elements = "abcd"
        for i in range(1000):
            # test random pairs of multisets
            p = Counter(dict((elem, randrange(-2, 4)) for elem in elements))
            p.update(e=1, f=-1, g=0)
            q = Counter(dict((elem, randrange(-2, 4)) for elem in elements))
            q.update(h=1, i=-1, j=0)
            for counterop, numberop in [
                (Counter.__add__, lambda x, y: max(0, x + y)),
                (Counter.__sub__, lambda x, y: max(0, x - y)),
                (Counter.__or__, lambda x, y: max(0, x, y)),
                (Counter.__and__, lambda x, y: max(0, min(x, y))),
            ]:
                result = counterop(p, q)
                for x in elements:
                    self.assertEqual(
                        numberop(p[x], q[x]), result[x], (counterop, x, p, q)
                    )
                # verify that results exclude non-positive counts
                self.assertTrue(x > 0 for x in result.values())

        elements = "abcdef"
        for i in range(100):
            # verify that random multisets with no repeats are exactly like
            # sets
            p = Counter(dict((elem, randrange(0, 2)) for elem in elements))
            q = Counter(dict((elem, randrange(0, 2)) for elem in elements))
            for counterop, setop in [
                (Counter.__sub__, set.__sub__),
                (Counter.__or__, set.__or__),
                (Counter.__and__, set.__and__),
            ]:
                counter_result = counterop(p, q)
                set_result = setop(set(p.elements()), set(q.elements()))
                self.assertEqual(counter_result, dict.fromkeys(set_result, 1))

    def test_inplace_operations(self):
        elements = "abcd"
        for i in range(1000):
            # test random pairs of multisets
            p = Counter(dict((elem, randrange(-2, 4)) for elem in elements))
            p.update(e=1, f=-1, g=0)
            q = Counter(dict((elem, randrange(-2, 4)) for elem in elements))
            q.update(h=1, i=-1, j=0)
            for inplace_op, regular_op in [
                (Counter.__iadd__, Counter.__add__),
                (Counter.__isub__, Counter.__sub__),
                (Counter.__ior__, Counter.__or__),
                (Counter.__iand__, Counter.__and__),
            ]:
                c = p.copy()
                c_id = id(c)
                regular_result = regular_op(c, q)
                inplace_result = inplace_op(c, q)
                self.assertEqual(inplace_result, regular_result)
                self.assertEqual(id(inplace_result), c_id)

    def test_subtract(self):
        c = Counter(a=-5, b=0, c=5, d=10, e=15, g=40)
        c.subtract(a=1, b=2, c=-3, d=10, e=20, f=30, h=-50)
        self.assertEqual(c, Counter(a=-6, b=-2, c=8, d=0, e=-5, f=-30, g=40, h=50))
        c = Counter(a=-5, b=0, c=5, d=10, e=15, g=40)
        c.subtract(Counter(a=1, b=2, c=-3, d=10, e=20, f=30, h=-50))
        self.assertEqual(c, Counter(a=-6, b=-2, c=8, d=0, e=-5, f=-30, g=40, h=50))
        c = Counter("aaabbcd")
        c.subtract("aaaabbcce")
        self.assertEqual(c, Counter(a=-1, b=0, c=-1, d=1, e=-1))

    def test_unary(self):
        c = Counter(a=-5, b=0, c=5, d=10, e=15, g=40)
        self.assertEqual(dict(+c), dict(c=5, d=10, e=15, g=40))
        self.assertEqual(dict(-c), dict(a=5))

    def test_repr_nonsortable(self):
        c = Counter(a=2, b=None)
        r = repr(c)
        self.assertIn("'a': 2", r)
        self.assertIn("'b': None", r)

    def test_helper_function(self):
        from xotl.tools.future.collections import _count_elements

        # two paths, one for real dicts and one for other mappings
        elems = list("abracadabra")

        d = dict()
        _count_elements(d, elems)
        self.assertEqual(d, {"a": 5, "r": 2, "b": 2, "c": 1, "d": 1})

        m = OrderedDict()
        _count_elements(m, elems)
        self.assertEqual(
            m, OrderedDict([("a", 5), ("b", 2), ("r", 2), ("c", 1), ("d", 1)])
        )


class TestOrderedDict(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(TypeError):
            # too many args
            OrderedDict([("a", 1), ("b", 2)], None)
        pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
        # dict input
        self.assertEqual(sorted(OrderedDict(dict(pairs)).items()), pairs)
        # kwds input
        self.assertEqual(sorted(OrderedDict(**dict(pairs)).items()), pairs)
        # pairs input
        self.assertEqual(list(OrderedDict(pairs).items()), pairs)
        # mixed input
        self.assertEqual(
            list(
                OrderedDict([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5).items()
            ),
            pairs,
        )

        # Make sure that direct calls to __init__ do not clear previous
        # contents
        d = OrderedDict([("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)])
        d.__init__([("e", 5), ("f", 6)], g=7, d=4)
        self.assertEqual(
            list(d.items()),
            [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5), ("f", 6), ("g", 7)],
        )

    def test_update(self):
        with self.assertRaises(TypeError):
            # too many args
            OrderedDict().update([("a", 1), ("b", 2)], None)
        pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
        od = OrderedDict()
        od.update(dict(pairs))
        # dict input
        self.assertEqual(sorted(od.items()), pairs)
        od = OrderedDict()
        od.update(**dict(pairs))
        # kwds input
        self.assertEqual(sorted(od.items()), pairs)
        od = OrderedDict()
        od.update(pairs)
        # pairs input
        self.assertEqual(list(od.items()), pairs)
        od = OrderedDict()
        od.update([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5)
        # mixed input
        self.assertEqual(list(od.items()), pairs)

        # Issue 9137: Named argument called 'other' or 'self'
        # shouldn't be treated specially.
        od = OrderedDict()
        od.update(self=23)
        self.assertEqual(list(od.items()), [("self", 23)])
        od = OrderedDict()
        od.update(other={})
        self.assertEqual(list(od.items()), [("other", {})])
        od = OrderedDict()
        od.update(red=5, blue=6, other=7, self=8)
        self.assertEqual(
            sorted(list(od.items())),
            [("blue", 6), ("other", 7), ("red", 5), ("self", 8)],
        )

        # Make sure that direct calls to update do not clear previous contents
        # add that updates items are not moved to the end
        d = OrderedDict([("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)])
        d.update([("e", 5), ("f", 6)], g=7, d=4)
        self.assertEqual(
            list(d.items()),
            [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5), ("f", 6), ("g", 7)],
        )

    def test_abc(self):
        self.assertIsInstance(OrderedDict(), MutableMapping)
        self.assertTrue(issubclass(OrderedDict, MutableMapping))

    def test_clear(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = OrderedDict(pairs)
        self.assertEqual(len(od), len(pairs))
        od.clear()
        self.assertEqual(len(od), 0)

    def test_delitem(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = OrderedDict(pairs)
        del od["a"]
        self.assertNotIn("a", od)
        with self.assertRaises(KeyError):
            del od["a"]
        self.assertEqual(list(od.items()), pairs[:2] + pairs[3:])

    def test_setitem(self):
        od = OrderedDict([("d", 1), ("b", 2), ("c", 3), ("a", 4), ("e", 5)])
        od["c"] = 10  # existing element
        od["f"] = 20  # new element
        self.assertEqual(
            list(od.items()),
            [("d", 1), ("b", 2), ("c", 10), ("a", 4), ("e", 5), ("f", 20)],
        )

    def test_iterators(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = OrderedDict(pairs)
        self.assertEqual(list(od), [t[0] for t in pairs])
        self.assertEqual(list(od.keys()), [t[0] for t in pairs])
        self.assertEqual(list(od.values()), [t[1] for t in pairs])
        self.assertEqual(list(od.items()), pairs)
        self.assertEqual(list(reversed(od)), [t[0] for t in reversed(pairs)])

    def test_popitem(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = OrderedDict(pairs)
        while pairs:
            self.assertEqual(od.popitem(), pairs.pop())
        with self.assertRaises(KeyError):
            od.popitem()
        self.assertEqual(len(od), 0)

    def test_pop(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = OrderedDict(pairs)
        shuffle(pairs)
        while pairs:
            k, v = pairs.pop()
            self.assertEqual(od.pop(k), v)
        with self.assertRaises(KeyError):
            od.pop("xyz")
        self.assertEqual(len(od), 0)
        self.assertEqual(od.pop(k, 12345), 12345)

        # make sure pop still works when __missing__ is defined
        class Missing(OrderedDict):
            def __missing__(self, key):
                return 0

        m = Missing(a=1)
        self.assertEqual(m.pop("b", 5), 5)
        self.assertEqual(m.pop("a", 6), 1)
        self.assertEqual(m.pop("a", 6), 6)
        with self.assertRaises(KeyError):
            m.pop("a")

    def test_equality(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od1 = OrderedDict(pairs)
        od2 = OrderedDict(pairs)
        # same order implies equality
        self.assertEqual(od1, od2)
        pairs = pairs[2:] + pairs[:2]
        od2 = OrderedDict(pairs)
        # different order implies inequality
        self.assertNotEqual(od1, od2)
        # comparison to regular dict is not order sensitive
        self.assertEqual(od1, dict(od2))
        self.assertEqual(dict(od2), od1)
        # different length implied inequality
        self.assertNotEqual(od1, OrderedDict(pairs[:-1]))

    def test_copying(self):
        # Check that ordered dicts are copyable, deepcopyable, picklable,
        # and have a repr/eval round-trip
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = OrderedDict(pairs)
        update_test = OrderedDict()
        update_test.update(od)
        for i, dup in enumerate(
            [
                od.copy(),
                copy.copy(od),
                copy.deepcopy(od),
                pickle.loads(pickle.dumps(od, 0)),
                pickle.loads(pickle.dumps(od, 1)),
                pickle.loads(pickle.dumps(od, 2)),
                pickle.loads(pickle.dumps(od, -1)),
                eval(repr(od)),
                update_test,
                OrderedDict(od),
            ]
        ):
            self.assertTrue(dup is not od)
            self.assertEqual(dup, od)
            self.assertEqual(list(dup.items()), list(od.items()))
            self.assertEqual(len(dup), len(od))
            self.assertEqual(type(dup), type(od))

    def test_yaml_linkage(self):
        # Verify that __reduce__ is setup in a way that supports PyYAML's
        # dump() feature.
        # In yaml, lists are native but tuples are not.
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = OrderedDict(pairs)
        # yaml.dump(od) -->
        # '!!python/object/apply:__main__.OrderedDict\n- - [a, 1]\n  - [b, 2]\n'
        self.assertTrue(all(type(pair) == list for pair in od.__reduce__()[1]))

    def test_reduce_not_too_fat(self):
        import sys

        # do not save instance dictionary if not needed
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = OrderedDict(pairs)
        self.assertIsNone(od.__reduce__()[2])
        od.x = 10
        self.assertIsNotNone(od.__reduce__()[2])

    def test_repr(self):
        od = OrderedDict([("c", 1), ("b", 2), ("a", 3)])
        self.assertEqual(repr(od), "OrderedDict([('c', 1), ('b', 2), ('a', 3)])")
        self.assertEqual(eval(repr(od)), od)
        self.assertEqual(repr(OrderedDict()), "OrderedDict()")

    def test_repr_recursive(self):
        # See issue #9826
        od = OrderedDict.fromkeys("abc")
        od["x"] = od
        self.assertEqual(
            repr(od),
            ("OrderedDict([('a', None), ('b', None), " "('c', None), ('x', ...)])"),
        )

    def test_setdefault(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = OrderedDict(pairs)
        pair_order = list(od.items())
        self.assertEqual(od.setdefault("a", 10), 3)
        # make sure order didn't change
        self.assertEqual(list(od.items()), pair_order)
        self.assertEqual(od.setdefault("x", 10), 10)
        # make sure 'x' is added to the end
        self.assertEqual(list(od.items())[-1], ("x", 10))

        # make sure setdefault still works when __missing__ is defined
        class Missing(OrderedDict):
            def __missing__(self, key):
                return 0

        self.assertEqual(Missing().setdefault(5, 9), 9)

    def test_reinsert(self):
        # Given insert a, insert b, delete a, re-insert a,
        # verify that a is now later than b.
        od = OrderedDict()
        od["a"] = 1
        od["b"] = 2
        del od["a"]
        od["a"] = 1
        self.assertEqual(list(od.items()), [("b", 2), ("a", 1)])

    def test_move_to_end(self):
        od = OrderedDict.fromkeys("abcde")
        self.assertEqual(list(od), list("abcde"))
        od.move_to_end("c")
        self.assertEqual(list(od), list("abdec"))
        od.move_to_end("c", 0)
        self.assertEqual(list(od), list("cabde"))
        od.move_to_end("c", 0)
        self.assertEqual(list(od), list("cabde"))
        od.move_to_end("e")
        self.assertEqual(list(od), list("cabde"))
        with self.assertRaises(KeyError):
            od.move_to_end("x")

    @unittest.skipIf("PyPy" in sys.version, "sys.getsizeof not supported")
    def test_sizeof(self):
        # Wimpy test: Just verify the reported size is larger than a regular
        # dict
        d = dict(a=1)
        od = OrderedDict(**d)
        self.assertGreater(sys.getsizeof(od), sys.getsizeof(d))

    def test_override_update(self):
        # Verify that subclasses can override update() without breaking
        # __init__()
        class MyOD(OrderedDict):
            def update(self, *args, **kwds):
                raise Exception()

        items = [("a", 1), ("c", 3), ("b", 2)]
        self.assertEqual(list(MyOD(items).items()), items)


class TestRankedDict(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(TypeError):
            # too many args
            RankedDict([("a", 1), ("b", 2)], None)
        pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
        # dict input
        self.assertEqual(sorted(RankedDict(dict(pairs)).items()), pairs)
        # kwds input
        self.assertEqual(sorted(RankedDict(**dict(pairs)).items()), pairs)
        # pairs input
        self.assertEqual(list(RankedDict(pairs).items()), pairs)
        # mixed input
        self.assertNotEqual(
            list(
                RankedDict([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5).items()
            ),
            pairs,
        )

        # Make sure that direct calls to __init__ do not clear previous
        # contents
        d = RankedDict([("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)])
        d.__init__([("f", 6), ("e", 5)], d=4)
        self.assertEqual(
            list(d.items()),
            [("a", 1), ("b", 2), ("c", 3), ("f", 6), ("e", 5), ("d", 4)],
        )

    def test_update(self):
        with self.assertRaises(TypeError):
            # too many args
            RankedDict().update([("a", 1), ("b", 2)], None)
        pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
        od = RankedDict()
        od.update(dict(pairs))
        # dict input
        self.assertEqual(sorted(od.items()), pairs)
        od = RankedDict()
        od.update(**dict(pairs))
        # kwds input
        self.assertEqual(sorted(od.items()), pairs)
        od = RankedDict()
        od.update(pairs)
        # pairs input
        self.assertEqual(list(od.items()), pairs)
        od = RankedDict()
        od.update([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5)
        # mixed input
        self.assertNotEqual(list(od.items()), pairs)

        # Issue 9137: Named argument called 'other' or 'self'
        # shouldn't be treated specially.
        od = RankedDict()
        od.update(self=23)
        self.assertEqual(list(od.items()), [("self", 23)])
        od = RankedDict()
        od.update(other={})
        self.assertEqual(list(od.items()), [("other", {})])
        od = RankedDict()
        od.update(red=5, blue=6, other=7, self=8)
        self.assertEqual(
            sorted(list(od.items())),
            [("blue", 6), ("other", 7), ("red", 5), ("self", 8)],
        )

    def test_abc(self):
        self.assertIsInstance(RankedDict(), MutableMapping)
        self.assertTrue(issubclass(RankedDict, MutableMapping))

    def test_clear(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        self.assertEqual(len(od), len(pairs))
        od.clear()
        self.assertEqual(len(od), 0)

    def test_delitem(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = RankedDict(pairs)
        del od["a"]
        self.assertNotIn("a", od)
        with self.assertRaises(KeyError):
            del od["a"]
        self.assertEqual(list(od.items()), pairs[:2] + pairs[3:])

    def test_setitem(self):
        od = RankedDict([("d", 1), ("b", 2), ("c", 3), ("a", 4), ("e", 5)])
        od["c"] = 10  # existing element
        od["f"] = 20  # new element
        self.assertEqual(
            list(od.items()),
            [("d", 1), ("b", 2), ("a", 4), ("e", 5), ("c", 10), ("f", 20)],
        )

    def test_iterators(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        self.assertEqual(list(od), [t[0] for t in pairs])
        self.assertEqual(list(od.keys()), [t[0] for t in pairs])
        self.assertEqual(list(od.values()), [t[1] for t in pairs])
        self.assertEqual(list(od.items()), pairs)
        self.assertEqual(list(reversed(od)), [t[0] for t in reversed(pairs)])

    def test_popitem(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        while pairs:
            self.assertEqual(od.popitem(), pairs.pop())
        with self.assertRaises(KeyError):
            od.popitem()
        self.assertEqual(len(od), 0)

    def test_pop(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        shuffle(pairs)
        while pairs:
            k, v = pairs.pop()
            self.assertEqual(od.pop(k), v)
        with self.assertRaises(KeyError):
            od.pop("xyz")
        self.assertEqual(len(od), 0)
        self.assertEqual(od.pop(k, 12345), 12345)

        # make sure pop still works when __missing__ is defined
        class Missing(RankedDict):
            def __missing__(self, key):
                return 0

        m = Missing(a=1)
        self.assertEqual(m.pop("b", 5), 5)
        self.assertEqual(m.pop("a", 6), 1)
        self.assertEqual(m.pop("a", 6), 6)
        with self.assertRaises(KeyError):
            m.pop("a")

    def test_equality(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od1 = RankedDict(pairs)
        od2 = RankedDict(pairs)
        # same order implies equality
        self.assertEqual(od1, od2)
        pairs = pairs[2:] + pairs[:2]
        od2 = RankedDict(pairs)
        # different order implies inequality
        self.assertNotEqual(od1, od2)
        # comparison to regular dict is not order sensitive
        self.assertEqual(od1, dict(od2))
        self.assertEqual(dict(od2), od1)
        # different length implied inequality
        self.assertNotEqual(od1, RankedDict(pairs[:-1]))

    def test_copying(self):
        # Check that ranked dicts are copyable, deepcopyable, picklable,
        # and have a repr/eval round-trip
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = RankedDict(pairs)
        update_test = RankedDict()
        update_test.update(od)
        for i, dup in enumerate(
            [
                od.copy(),
                copy.copy(od),
                copy.deepcopy(od),
                pickle.loads(pickle.dumps(od, 0)),
                pickle.loads(pickle.dumps(od, 1)),
                pickle.loads(pickle.dumps(od, 2)),
                pickle.loads(pickle.dumps(od, -1)),
                eval(repr(od)),
                update_test,
                RankedDict(od),
            ]
        ):
            self.assertTrue(dup is not od)
            self.assertEqual(dup, od)
            self.assertEqual(list(dup.items()), list(od.items()))
            self.assertEqual(len(dup), len(od))
            self.assertEqual(type(dup), type(od))

    def test_yaml_linkage(self):
        # Verify that __reduce__ is setup in a way that supports PyYAML's
        # dump() feature.
        # In yaml, lists are native but tuples are not.
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = RankedDict(pairs)
        # yaml.dump(od) -->
        # '!!python/object/apply:__main__.RankedDict\n- - [a, 1]\n  - [b, 2]\n'
        self.assertTrue(all(type(pair) == list for pair in od.__reduce__()[1]))

    def test_repr(self):
        od = RankedDict([("c", 1), ("b", 2), ("a", 3)])
        self.assertEqual(repr(od), "RankedDict([('c', 1), ('b', 2), ('a', 3)])")
        self.assertEqual(eval(repr(od)), od)
        self.assertEqual(repr(RankedDict()), "RankedDict()")

    def test_repr_recursive(self):
        # See issue #9826
        od = RankedDict.fromkeys("abc")
        od["x"] = od
        self.assertEqual(
            repr(od),
            ("RankedDict([('a', None), ('b', None), " "('c', None), ('x', ...)])"),
        )

    def test_setdefault(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        pair_order = list(od.items())
        self.assertEqual(od.setdefault("a", 10), 3)
        # make sure order didn't change
        self.assertEqual(list(od.items()), pair_order)
        self.assertEqual(od.setdefault("x", 10), 10)
        # make sure 'x' is added to the end
        self.assertEqual(list(od.items())[-1], ("x", 10))

        # make sure setdefault still works when __missing__ is defined
        class Missing(RankedDict):
            def __missing__(self, key):
                return 0

        self.assertEqual(Missing().setdefault(5, 9), 9)

    def test_reinsert(self):
        # Given insert a, insert b, delete a, re-insert a,
        # verify that a is now later than b.
        od = RankedDict()
        od["a"] = 1
        od["b"] = 2
        del od["a"]
        od["a"] = 1
        self.assertEqual(list(od.items()), [("b", 2), ("a", 1)])

    def test_move_to_end(self):
        od = RankedDict.fromkeys("abcde")
        self.assertEqual(list(od), list("abcde"))
        od.move_to_end("c")
        self.assertEqual(list(od), list("abdec"))
        od.move_to_end("c", 0)
        self.assertEqual(list(od), list("cabde"))
        od.move_to_end("c", 0)
        self.assertEqual(list(od), list("cabde"))
        od.move_to_end("e")
        self.assertEqual(list(od), list("cabde"))
        with self.assertRaises(KeyError):
            od.move_to_end("x")

    @unittest.skipIf("PyPy" in sys.version, "sys.getsizeof not supported")
    def test_sizeof(self):
        # Wimpy test: Just verify the reported size is larger than a regular
        # dict
        d = dict(a=1)
        od = RankedDict(**d)
        self.assertGreater(sys.getsizeof(od), sys.getsizeof(d))


class TestPascalSet(unittest.TestCase):
    def test_consistency(self):
        from random import randint
        from xotl.tools.future.collections import PascalSet

        count = 5
        for test in range(count):
            size = randint(20, 60)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s1 = PascalSet(*ranges)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s2 = PascalSet(*ranges)
            ss1 = set(s1)
            ss2 = set(s2)
            self.assertEqual(s1, ss1)
            self.assertEqual(s1 - s2, ss1 - ss2)
            self.assertEqual(s2 - s1, ss2 - ss1)
            self.assertEqual(s1 & s2, ss1 & ss2)
            self.assertEqual(s2 & s1, ss2 & ss1)
            self.assertEqual(s1 | s2, ss1 | ss2)
            self.assertEqual(s2 | s1, ss2 | ss1)
            self.assertEqual(s1 ^ s2, ss1 ^ ss2)
            self.assertEqual(s2 ^ s1, ss2 ^ ss1)
            self.assertLess(s1 - s2, s1)
            self.assertLess(s1 - s2, ss1)
            self.assertLessEqual(s1 - s2, s1)
            self.assertLessEqual(s1 - s2, ss1)
            self.assertGreater(s1, s1 - s2)
            self.assertGreater(s1, ss1 - ss2)
            self.assertGreaterEqual(s1, s1 - s2)
            self.assertGreaterEqual(s1, ss1 - ss2)

    def test_syntax_sugar(self):
        from xotl.tools.future.collections import PascalSet

        s1 = PascalSet[1:4, 9, 15:18]
        s2 = PascalSet[3:18]
        self.assertEqual(str(s1), "{1..3, 9, 15..17}")
        self.assertEqual(str(s1 ^ s2), "{1, 2, 4..8, 10..14}")
        self.assertEqual(list(PascalSet[3:18]), list(range(3, 18)))

    def test_operators(self):
        from xotl.tools.future.collections import PascalSet

        g = lambda s: (i for i in s)
        s1 = PascalSet[1:4, 9, 15:18]
        r1 = range(1, 18)
        s2 = PascalSet(s1, 20)
        self.assertTrue(s1.issubset(s1))
        self.assertTrue(s1.issubset(set(s1)))
        self.assertTrue(s1.issubset(list(s1)))
        self.assertTrue(s1.issubset(g(s1)))
        self.assertTrue(s1.issubset(r1))
        self.assertTrue(s1.issubset(set(r1)))
        self.assertTrue(s1.issubset(list(r1)))
        self.assertTrue(s1.issubset(g(r1)))
        self.assertTrue(s2.issuperset(s2))
        self.assertTrue(s2.issuperset(s1))
        self.assertTrue(s2.issuperset(set(s1)))
        self.assertTrue(s2.issuperset(list(s1)))
        self.assertTrue(s2.issuperset(g(s1)))
        self.assertTrue(s1 <= set(s1))
        self.assertTrue(s1 < s2)
        self.assertTrue(s1 <= s2)
        self.assertTrue(s1 < set(s2))
        self.assertTrue(s1 <= set(s2))
        self.assertTrue(s1 < set(r1))
        self.assertTrue(s1 <= set(r1))
        self.assertTrue(s2 >= s2)
        self.assertTrue(s2 >= set(s2))
        self.assertTrue(s2 > s1)
        self.assertTrue(s2 > set(s1))
        self.assertTrue(s2 >= s1)
        self.assertTrue(s2 >= set(s1))

    def test_errors(self):
        """Test that stacked.pop has the same semantics has dict.pop."""
        from xotl.tools.future.collections import PascalSet

        s1 = PascalSet[1:4, 9, 15:18]
        s2 = PascalSet(s1, 20)
        self.assertLess(s1, s2)
        try:
            if s1 < list(s2):
                state = "less"
            else:
                state = "not-less"
        except TypeError:
            state = "TypeError"
        self.assertEqual(state, "TypeError")
        with self.assertRaises(TypeError):
            if s1 < set(s2):
                state = "ok"
            if s1 < list(s2):
                state = "safe-less"
            else:
                state = "safe-not-less"
        self.assertEqual(state, "ok")


class TestBitPascalSet(unittest.TestCase):
    def test_consistency(self):
        from random import randint
        from xotl.tools.future.collections import BitPascalSet

        count = 5
        for test in range(count):
            size = randint(20, 60)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s1 = BitPascalSet(*ranges)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s2 = BitPascalSet(*ranges)
            ss1 = set(s1)
            ss2 = set(s2)
            self.assertEqual(s1, ss1)
            self.assertEqual(s1 - s2, ss1 - ss2)
            self.assertEqual(s2 - s1, ss2 - ss1)
            self.assertEqual(s1 & s2, ss1 & ss2)
            self.assertEqual(s2 & s1, ss2 & ss1)
            self.assertEqual(s1 | s2, ss1 | ss2)
            self.assertEqual(s2 | s1, ss2 | ss1)
            self.assertEqual(s1 ^ s2, ss1 ^ ss2)
            self.assertEqual(s2 ^ s1, ss2 ^ ss1)
            self.assertLess(s1 - s2, s1)
            self.assertLess(s1 - s2, ss1)
            self.assertLessEqual(s1 - s2, s1)
            self.assertLessEqual(s1 - s2, ss1)
            self.assertGreater(s1, s1 - s2)
            self.assertGreater(s1, ss1 - ss2)
            self.assertGreaterEqual(s1, s1 - s2)
            self.assertGreaterEqual(s1, ss1 - ss2)

    def test_syntax_sugar(self):
        from xotl.tools.future.collections import BitPascalSet

        s1 = BitPascalSet[1:4, 9, 15:18]
        s2 = BitPascalSet[3:18]
        self.assertEqual(str(s1), "{1..3, 9, 15..17}")
        self.assertEqual(str(s1 ^ s2), "{1, 2, 4..8, 10..14}")
        self.assertEqual(list(BitPascalSet[3:18]), list(range(3, 18)))

    def test_operators(self):
        from xotl.tools.future.collections import BitPascalSet

        g = lambda s: (i for i in s)
        s1 = BitPascalSet[1:4, 9, 15:18]
        r1 = range(1, 18)
        s2 = BitPascalSet(s1, 20)
        self.assertTrue(s1.issubset(s1))
        self.assertTrue(s1.issubset(set(s1)))
        self.assertTrue(s1.issubset(list(s1)))
        self.assertTrue(s1.issubset(g(s1)))
        self.assertTrue(s1.issubset(r1))
        self.assertTrue(s1.issubset(set(r1)))
        self.assertTrue(s1.issubset(list(r1)))
        self.assertTrue(s1.issubset(g(r1)))
        self.assertTrue(s2.issuperset(s2))
        self.assertTrue(s2.issuperset(s1))
        self.assertTrue(s2.issuperset(set(s1)))
        self.assertTrue(s2.issuperset(list(s1)))
        self.assertTrue(s2.issuperset(g(s1)))
        self.assertTrue(s1 <= set(s1))
        self.assertTrue(s1 < s2)
        self.assertTrue(s1 <= s2)
        self.assertTrue(s1 < set(s2))
        self.assertTrue(s1 <= set(s2))
        self.assertTrue(s1 < set(r1))
        self.assertTrue(s1 <= set(r1))
        self.assertTrue(s2 >= s2)
        self.assertTrue(s2 >= set(s2))
        self.assertTrue(s2 > s1)
        self.assertTrue(s2 > set(s1))
        self.assertTrue(s2 >= s1)
        self.assertTrue(s2 >= set(s1))

    def test_errors(self):
        """Test that stacked.pop has the same semantics has dict.pop."""
        from xotl.tools.future.collections import BitPascalSet

        s1 = BitPascalSet[1:4, 9, 15:18]
        s2 = BitPascalSet(s1, 20)
        self.assertLess(s1, s2)
        try:
            if s1 < list(s2):
                state = "less"
            else:
                state = "not-less"
        except TypeError:
            state = "TypeError"
        self.assertEqual(state, "TypeError")
        with self.assertRaises(TypeError):
            if s1 < set(s2):
                state = "ok"
            if s1 < list(s2):
                state = "safe-less"
            else:
                state = "safe-not-less"
        self.assertEqual(state, "ok")


class TestCodeDict(unittest.TestCase):
    def test_formatter(self):
        from xotl.tools.future.collections import codedict

        cd = codedict(x=1, y=2, z=3.0)
        self.assertEqual(
            "{_[x + y]} is 3 --  {_[x + z]} is 4.0".format(_=cd),
            "3 is 3 --  4.0 is 4.0",
        )
        self.assertEqual(
            cd >> "{_[x + y]} is 3 --  {_[x + z]} is 4.0 -- {x} is 1",
            "3 is 3 --  4.0 is 4.0 -- 1 is 1",
        )
        self.assertEqual(
            "{_[x + y]} is 3 --  {_[x + z]} is 4.0 -- {x} is 1" << cd,
            "3 is 3 --  4.0 is 4.0 -- 1 is 1",
        )


def test_abcs():
    from xotl.tools.future.collections import Container  # noqa
    from xotl.tools.future.collections import Iterable  # noqa
    from xotl.tools.future.collections import Iterator  # noqa
    from xotl.tools.future.collections import Sized  # noqa
    from xotl.tools.future.collections import Callable  # noqa
    from xotl.tools.future.collections import Sequence  # noqa
    from xotl.tools.future.collections import MutableSequence  # noqa
    from xotl.tools.future.collections import Set  # noqa
    from xotl.tools.future.collections import MutableSet  # noqa
    from xotl.tools.future.collections import Mapping  # noqa
    from xotl.tools.future.collections import MutableMapping  # noqa
    from xotl.tools.future.collections import MappingView  # noqa
    from xotl.tools.future.collections import ItemsView  # noqa
    from xotl.tools.future.collections import KeysView  # noqa
    from xotl.tools.future.collections import ValuesView  # noqa


def test_opendict():
    try:
        from enum import Enum
    except ImportError:
        from enum34 import Enum

    from xotl.tools.future.collections import opendict

    class Foo:
        x = 1
        _y = 2

    foo = opendict.from_enum(Foo)
    assert dict(foo) == {"x": 1}

    class Bar(Enum):
        spam = "spam"

        def eat(self):
            return self.spam

    bar = opendict.from_enum(Bar)
    assert dict(bar) == {"spam": Bar.spam}


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main(verbosity=2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest
import pytest

# Test concurrent access to context by several greenlets.  Verify isolation in
# the greenlets.  We don't test isolation for threads cause that depends on
# python's thread locals and we *rely* on its correctness.
#
# Since xotl.tools.context inspects sys.modules to test for greenlet presence
# we need to import greenlets before importing context.
#
try:
    import greenlet
except ImportError:
    GREENLETS = False
else:
    GREENLETS = True

import sys

sys.modules.pop("xotl.tools.tasking", None)
sys.modules.pop("xotl.tools.context", None)
del sys

from xotl.tools.context import context


class TestContext(unittest.TestCase):
    def test_simple_contexts(self):
        with context("CONTEXT-1"):
            self.assertIsNot(None, context["CONTEXT-1"])
            with context("CONTEXT-1"):
                with context("context-2"):
                    self.assertIsNot(None, context["CONTEXT-1"])
                    self.assertIsNot(None, context["context-2"])
                self.assertEqual(False, bool(context["context-2"]))
            self.assertIsNot(None, context["CONTEXT-1"])
        self.assertEqual(False, bool(context["CONTEXT-1"]))

    def test_with_objects(self):
        CONTEXT1 = object()
        CONTEXT2 = object()
        with context(CONTEXT1):
            self.assertIsNot(None, context[CONTEXT1])
            with context(CONTEXT1):
                with context(CONTEXT2):
                    self.assertIsNot(None, context[CONTEXT1])
                    self.assertIsNot(None, context[CONTEXT2])
                self.assertEqual(False, bool(context[CONTEXT2]))
            self.assertIsNot(None, context[CONTEXT1])
        self.assertEqual(False, bool(context[CONTEXT1]))


def test_stacking_of_data_does_not_leak():
    c1 = "CONTEXT-1"
    with context(c1, a=1, b=1) as cc1:
        assert cc1["a"] == 1
        with context(c1, a=2, z="zzz") as cc2:
            assert cc2 is cc1
            assert cc2["a"] == 2
            assert cc2["b"] == 1  # Given by the upper enclosing level
            assert cc2["z"] == "zzz"

            # Let's change it for this level
            cc2["b"] = "jailed!"
            assert cc2["b"] == "jailed!"

        # But in the upper level both a and b stay the same
        assert cc1["a"] == 1
        assert cc1["b"] == 1
        assert set(cc1) == {"a", "b"}

    try:
        assert cc1["a"] == 1
        assert False
    except (IndexError, KeyError):
        pass


def test_data_is_an_opendict():
    c1 = object()
    with context(c1, a=1, b=1) as cc1:
        with context(c1, a=2) as cc2:
            assert cc2 is cc1
            assert cc2.a == 2
            assert cc2.b == 1  # Given by the upper enclosing level
            cc2.b = "jaile!d"
        assert cc1.a == 1
        assert cc1["b"] == 1


def test_reusing_raises():
    with context("a") as a:
        try:
            with a:
                pass
            assert False, "It should have raised a RuntimeError"
        except RuntimeError:
            pass
        except:  # noqa
            assert False, "It should have raised a RuntimeError"


def test_from_dicts():
    with context.from_dicts("A", dict(a=1), dict(a=2, b=1)) as c:
        assert c["a"] == 1
        assert c["b"] == 1
        with context.from_dicts("A", dict(a=2), dict(b=2)) as c:
            assert c["b"] == 1
            assert c["a"] == 2
        assert c["b"] == 1
        assert c["a"] == 1


def test_from_defaults():
    with context.from_defaults("A", a=1):
        with context.from_defaults("A", a=2, b=1) as c:
            assert c["a"] == 1
            assert c["b"] == 1
            with context("A", a=2) as c2:
                assert c2["a"] == 2
            # It recovers the value
            assert c["a"] == 1
        # and again
        assert c["a"] == 1


def test_recover_from_runtime_bug_33():
    try:
        with context("A") as c:
            with c:
                pass
    except RuntimeError:
        pass

    with context("A"):
        pass


def test_null_context_is_mapping():
    from xotl.tools.context import NullContext

    dict(**NullContext())


@pytest.mark.skipif(not GREENLETS, reason="greenlet is not installed")
def test_greenlet_contexts():
    import random
    from xotl.tools.symbols import Unset

    calls = 0
    switches = 0

    class GreenletProg:
        def __init__(self, arg):
            self.arg = arg

        def __call__(self):
            nonlocal calls
            nonlocal switches
            calls += 1
            switches += 1
            assert "GREEN CONTEXT" not in context
            with context("GREEN CONTEXT") as ctx:
                assert ctx.get("greenvar", Unset) is Unset
                ctx["greenvar"] = self.arg
                root.switch()
                switches += 1
                assert ctx["greenvar"] == self.arg
                # list() makes KeyViews pass in Python 3+
                assert list(ctx.keys()) == ["greenvar"]

    def loop(n):
        nonlocal calls
        nonlocal switches
        greenlets = [greenlet.greenlet(run=GreenletProg(i)) for i in range(n)]
        calls = 0
        switches = 0
        while greenlets:
            pos = random.randrange(0, len(greenlets))
            gl = greenlets[pos]
            gl.switch()
            # The gl has relinquished control, so if its dead removed from the
            # list, otherwise let it be for another round.
            if gl.dead:
                del greenlets[pos]
        assert calls == n, "There should be N calls to greenlets."
        assert switches == 2 * n, "There should be 2*N switches."

    def loop_determ(n):
        nonlocal calls
        nonlocal switches
        greenlets = [greenlet.greenlet(run=GreenletProg(i)) for i in range(n)]
        pos = 0
        calls = 0
        switches = 0
        while greenlets:
            gl = greenlets[pos]
            gl.switch()
            # The gl has relinquished control, so if its dead removed from the
            # list, otherwise let it be for another round.
            if gl.dead:
                del greenlets[pos]
            # In this case we ensure there will be several concurrent
            # greenlets
            pos = ((pos + 1) % len(greenlets)) if greenlets else 0
        assert calls == n, "There should be N calls to greenlets."
        assert switches == 2 * n, "There should be 2*N switches."

    root = greenlet.greenlet(run=loop)
    root.switch(10)

    root = greenlet.greenlet(run=loop_determ)
    root.switch(5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from hypothesis import given
from hypothesis.strategies import text, binary


@given(s=binary())
def test_safe_decode_dont_fail_uppon_invalid_encoding(s):
    from xotl.tools.future.codecs import safe_decode

    assert safe_decode(s, "i-dont-exist") == safe_decode(s)


@given(s=text())
def test_safe_encode_dont_fail_uppon_invalid_encoding(s):
    from xotl.tools.future.codecs import safe_encode

    assert safe_encode(s, "i-dont-exist") == safe_encode(s)


@given(text())
def test_safe_encode_yields_bytes(s):
    from xotl.tools.future.codecs import safe_encode

    assert isinstance(safe_encode(s), bytes)


@given(binary())
def test_safe_decode_yields_unicode(s):
    try:
        Text = unicode
    except NameError:
        Text = str  # Python 3
    from xotl.tools.future.codecs import safe_decode

    assert isinstance(safe_decode(s), Text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import pytest

from xotl.tools.objects import (
    fulldir,
    smart_copy,
    lazy,
    setdefaultattr,
    traverse,
    get_traverser,
    dict_merge,
    get_first_of,
    smart_getter,
    smart_setter,
    extract_attrs,
    copy_class,
    validate_attrs,
    memoized_property,
    classproperty,
    xproperty,
    staticproperty,
    multi_getter,
    save_attributes,
)
from xotl.tools.symbols import Unset
from xotl.tools.future.types import SimpleNamespace as new


def test_smart_copy():
    source = new(a=1, b=2, c=4, _d=5)
    target = {}
    smart_copy(source, target, defaults=False)
    assert target == dict(a=1, b=2, c=4)

    source = new(a=1, b=2, c=4, _d=5)
    target = {}
    smart_copy(source, target, defaults=None)
    assert target == dict(a=1, b=2, c=4)

    target = {}
    smart_copy(source, target, defaults=True)
    assert target["_d"] == 5


def test_smart_copy_with_defaults():
    defaults = {
        "host": "localhost",
        "port": 5432,
        "user": "openerp",
        "password": (KeyError, "{key}"),
    }
    kwargs = {"password": "keep-out!"}
    args = smart_copy(kwargs, {}, defaults=defaults)
    assert args == dict(
        host="localhost", port=5432, user="openerp", password="keep-out!"
    )

    # if missing a required key
    with pytest.raises(KeyError):
        args = smart_copy({}, {}, defaults=defaults)


def test_smart_copy_signature():
    with pytest.raises(TypeError):
        smart_copy({}, defaults=False)


def test_smart_copy_from_dict_to_dict():
    c = dict(c=1, d=23)
    d = dict(d=1)
    smart_copy(c, d)
    assert d == dict(c=1, d=23)


def test_smart_copy_with_plain_defaults():
    c = dict(a=1, b=2, c=3)
    d = {}
    smart_copy(c, d, defaults=("a", "x"))
    assert d == dict(a=1, x=None)


def test_smart_copy_with_callable_default():
    def default(attr, source=None):
        return attr in ("a", "b")

    c = dict(a=1, b="2", c="3x")
    d = {}
    smart_copy(c, d, defaults=default)
    assert d == dict(a=1, b="2")

    class inset:
        def __init__(self, items):
            self.items = items

        def __call__(self, attr, source=None):
            return attr in self.items

    c = dict(a=1, b="2", c="3x")
    d = {}
    smart_copy(c, d, defaults=inset("ab"))
    assert d == dict(a=1, b="2")


def test_fulldir():
    assert {"__getitem__", "get", "items", "keys"} < fulldir({})


def test_newstyle_metaclass():
    class Field:
        __slots__ = (str("name"), str("default"))

        def __init__(self, default):
            self.default = default

        def __get__(self, inst, owner):
            if not inst:
                return self
            return self.default

    class ModelType(type):
        pass

    class Base:
        def __init__(self, **attrs):
            self.__dict__.update(attrs)

    class Model(metaclass=ModelType):
        f1 = Field(1009)
        f2 = 0

        def __init__(self, **attrs):
            self.__dict__.update(attrs)

    class Model2(Base, metaclass=ModelType):
        pass

    class SubMeta(ModelType):
        pass

    class Submodel(Model, metaclass=SubMeta):
        pass

    inst = Model(name="Instance")
    assert inst.f1 == 1009
    assert inst.name == "Instance"
    assert isinstance(Model.f1, Field)
    assert type(Model) is ModelType
    assert type(Submodel) is SubMeta
    assert type(Model2) is ModelType
    assert Model2.__base__ is Base
    assert Submodel.__base__ is Model
    assert Model.__base__ is object


def test_new_style_metaclass_registration():
    class BaseMeta(type):
        classes = []

        def __new__(cls, name, bases, attrs):
            res = super(BaseMeta, cls).__new__(cls, name, bases, attrs)
            cls.classes.append(res)  # <-- side effect
            return res

    class Base(metaclass=BaseMeta):
        pass

    class SubType(BaseMeta):
        pass

    class Egg(Base, metaclass=SubType):
        pass

    assert Egg.__base__ is Base
    assert len(BaseMeta.classes) == 2

    class Spam(Base, metaclass=SubType):
        'Like "Egg" but it will be registered twice in Python 2.x.'

    assert len(BaseMeta.classes) == 3  # Properly called once in Python 3

    # Nevertheless the bases are ok.
    assert Spam.__bases__ == (Base,)


def test_lazy():
    class new:
        pass

    inst = new()
    setter = lambda a: -a
    setdefaultattr(inst, "c", lazy(setter, 10))
    assert inst.c == -10
    setdefaultattr(inst, "c", lazy(setter, 20))
    assert inst.c == -10


# Easly creates a hierarchy of objects
class New:
    def __init__(self, **kwargs):
        attrs = {}
        children = {}
        for attr, value in kwargs.items():
            if "." in attr:
                name, childattr = attr.split(".", 1)
                child = children.setdefault(name, {})
                child[childattr] = value
            else:
                attrs[attr] = value
        self.__dict__.update(attrs)
        assert set(attrs.keys()) & set(children.keys()) == set()
        for child, vals in children.items():
            setattr(self, child, New(**vals))


def test_traversing():
    obj = new(**{"a": 1, "b": new(c=new(d={"x": 2}, x=3))})
    assert traverse(obj, "a") == 1
    assert traverse(obj, "b.c.d.x") == 2
    assert traverse(obj, "b.c.x") == 3
    with pytest.raises(AttributeError):
        traverse(obj, "a.v")
    with pytest.raises(AttributeError):
        traverse(obj, "a.b.c.d.y")

    traverser = get_traverser("a", "b.c.d.x", "b.c.d.y")
    with pytest.raises(AttributeError):
        traverser(obj)
    obj.b.c.d["y"] = None
    assert traverser(obj) == (1, 2, None)


def test_traversing_bug_ignoring_getter():
    from unittest import mock
    from xoutil.objects import traverse

    sentinel = object()

    class Me:
        def __getattr__(self, attr):
            return self

    return_sentinel = mock.Mock(return_value=sentinel)

    me = Me()
    assert traverse(me, "x.y", getter=return_sentinel) is sentinel
    assert return_sentinel.called


def test_dict_merge_base_cases():
    base = {"a": "a", "d": {"attr1": 2}}
    assert dict_merge() == {}
    assert dict_merge(base) == base
    assert dict_merge(**base) == base


def test_dict_merge_simple_cases():
    first = {"a": {"attr1": 1}, "b": {"attr1": 1}, "c": 194, "shared": 1}
    second = {"a": {"attr2": 2}, "b": {"attr2": 2}, "d": 195, "shared": 2}
    expected = {
        "a": {"attr1": 1, "attr2": 2},
        "b": {"attr1": 1, "attr2": 2},
        "c": 194,
        "d": 195,
        "shared": 2,
    }
    assert dict_merge(first, second) == expected
    assert dict_merge(first, **second) == expected
    assert dict_merge(second, first) == dict(expected, shared=1)
    assert dict_merge(second, **first) == dict(expected, shared=1)


def test_dict_merge_compatible_cases():
    first = {192: ["attr1", 1], 193: {"attr1", 1}}
    second = {192: ("attr2", 2), 193: ["attr2", 2]}
    assert dict_merge(first, second) == {
        192: ["attr1", 1, "attr2", 2],
        193: {"attr1", 1, "attr2", 2},
    }
    result = dict_merge(second, first)
    assert result[192] == ("attr2", 2, "attr1", 1)
    key_193 = result[193]
    assert key_193[:2] == ["attr2", 2]
    # Since order of set's members is not defined we can't test order, we can
    # only know that they'll be in the last two positions.
    assert key_193.index("attr1") in (2, 3)
    assert key_193.index(1) in (2, 3)


def test_dict_merge_errors():
    from xoutil.objects import dict_merge

    first = {192: 192}
    second = {192: [192]}
    with pytest.raises(TypeError):
        dict_merge(second, first)
    with pytest.raises(TypeError):
        dict_merge(first, second)


def test_get_first_of():
    somedict = {"foo": "bar", "spam": "eggs"}
    assert get_first_of(somedict, "no", "foo", "spam") == "bar"

    somedict = {"foo": "bar", "spam": "eggs"}
    assert get_first_of(somedict, "eggs") is None

    class Someobject:
        pass

    inst = Someobject()
    inst.foo = "bar"
    inst.eggs = "spam"
    assert get_first_of(inst, "no", "eggs", "foo") == "spam"
    assert get_first_of(inst, "invalid") is None

    somedict = {"foo": "bar", "spam": "eggs"}

    class Someobject:
        pass

    inst = Someobject()
    inst.foo = "bar2"
    inst.eggs = "spam"
    assert get_first_of((somedict, inst), "eggs") == "spam"
    assert get_first_of((somedict, inst), "foo") == "bar"
    assert get_first_of((inst, somedict), "foo") == "bar2"
    assert get_first_of((inst, somedict), "foobar") is None

    none = object()
    assert get_first_of((inst, somedict), "foobar", default=none) is none
    _eggs = get_first_of(somedict, "foo", "spam", pred=lambda v: len(v) > 3)
    assert _eggs == "eggs"
    _none = get_first_of(somedict, "foo", "spam", pred=lambda v: len(v) > 4)
    assert _none is None

    with pytest.raises(TypeError):
        get_first_of(None, anything=1)


def test_smart_getter():
    o = new()
    o.attr1 = 1
    o.attr2 = 1
    getter = smart_getter(o)
    assert getter("attr1") == getter("attr2") == 1
    assert getter("attr3") is None

    getter = smart_getter(o, strict=True)
    assert getter("attr1") == getter("attr2") == 1
    with pytest.raises(AttributeError):
        assert getter("attr3") is None

    d = {"key1": 1, "key2": 1}
    getter = smart_getter(d)
    assert getter("key1") == getter("key2") == 1
    assert getter("key3") is None

    getter = smart_getter(d, strict=True)
    assert getter("key1") == getter("key2") == 1
    with pytest.raises(KeyError):
        assert getter("key3") is None
    assert getter("key3", None) is None


def test_smart_setter():
    class new:
        pass

    o = new()
    setter = smart_setter(o)
    setter("attr1", 1)
    setter("attr2", 1)
    assert o.attr1 == o.attr2 == 1

    d = {"key1": 1, "key2": 1}
    setter = smart_setter(d)
    setter("key1", 10)
    assert d["key1"] == 10


def test_extract_attrs():
    d = dict(a=(1,), b=2, c=3, x=4)
    assert extract_attrs(d, "a") == (1,)
    assert extract_attrs(d, "a", "b", "c", "x") == ((1,), 2, 3, 4)

    with pytest.raises(AttributeError):
        assert extract_attrs(d, "y")
    assert extract_attrs(d, "y", default=None) is None

    class new:
        def __init__(self, **kw):
            self.__dict__.update(kw)

    d = new(a=(1,), b=2, c=3, x=4)
    assert extract_attrs(d, "a") == (1,)
    assert extract_attrs(d, "a", "b", "c", "x") == ((1,), 2, 3, 4)

    with pytest.raises(AttributeError):
        assert extract_attrs(d, "y")
    assert extract_attrs(d, "y", default=None) is None


def test_copy_class():
    u = str

    class MetaFoo(type):
        pass

    class Foo(metaclass=MetaFoo):
        a = 1
        b = 2
        c = 3
        d = 4

    class Baz(Foo):
        e = 5

    index = {k: getattr(Foo, k) for k in "abcd"}
    Bar = copy_class(Foo)
    assert Bar.a == Foo.a and Bar.b and Bar.c and Bar.d

    Egg = copy_class(Foo, ignores=["b", "c"])
    assert getattr(Egg, "b", Unset) is Unset

    Egg = copy_class(Foo, ignores=[lambda k: index.get(k) and index.get(k) > 2])
    assert Egg.a == Foo.a
    assert getattr(Egg, "c", Unset) is Unset

    Named = copy_class(Foo, new_name="Named")
    assert Named.__name__ == "Named"

    Named = copy_class(Foo, new_name=u("Named"))
    assert Named.__name__ == "Named"

    import fnmatch

    pattern = lambda attr: fnmatch.fnmatch(attr, "a*")
    Egg = copy_class(Foo, ignores=[pattern])
    assert getattr(Egg, "a", Unset) is Unset

    import re

    _pattern = re.compile("^a")
    pattern = lambda attr: _pattern.match(attr)
    Egg = copy_class(Foo, ignores=[pattern])
    assert getattr(Egg, "a", Unset) is Unset


def test_validate_attrs():
    class Person:
        def __init__(self, **kwargs):
            for which in kwargs:
                setattr(self, which, kwargs[which])

    source = Person(name="Manuel", age=33, sex="male")
    target = {"name": "Manuel", "age": 4, "sex": "male"}

    assert validate_attrs(
        source, target, force_equals=("sex",), force_differents=("age",)
    )

    assert not validate_attrs(source, target, force_equals=("age",))


@pytest.mark.xfail()
def test_memoized_classproperty():
    current = 1

    class Foobar:
        @memoized_property
        @classproperty
        def prop(cls):
            return current

        @classproperty
        @memoized_property
        def prop2(cls):
            return current

    assert Foobar.prop == current
    current += 1
    assert Foobar.prop != current


def test_properties():
    _x = "static"

    class Foobar:
        _x = "class"

        def __init__(self):
            self._x = "instance"

        @xproperty
        def x(self):
            return self._x

        @classproperty
        def cprop(cls):
            return cls._x

        @staticproperty
        def sprop():
            return _x

    f = Foobar()

    assert Foobar.x == "class"
    assert f.x == "instance"

    assert Foobar.cprop == "class"
    assert f.cprop == "class"

    assert Foobar.sprop == "static"
    assert f.sprop == "static"


def test_multi_getter_failure():
    """`multi_getter` is not the same as `traverse`.

    When a collection of identifiers is given, it get the first valid value
    (see the documentation).

    """
    from xoutil.objects import traverse

    class new:
        def __init__(self, **k):
            self.__dict__.update(k)

    top = new(d=dict(a=1, b=2), a=10, b=20)

    assert traverse(top, "d.a") == 1
    assert next(multi_getter(top, ("d", "a"))) == {"a": 1, "b": 2}


def test_save_attributes():
    from xoutil.future.types import SimpleNamespace as new
    from xoutil.objects import save_attributes

    obj = new(a=1, b=2)
    with save_attributes(obj, "a"):
        obj.a = 2
        obj.b = 3
        assert obj.a == 2

    assert obj.a == 1
    assert obj.b == 3


def test_temp_attributes():
    from xotl.tools.future.types import SimpleNamespace as new
    from xotl.tools.objects import temp_attributes

    obj = new(a=1, b=2)
    with temp_attributes(obj, dict(a=2)):
        assert obj.a == 2
        assert obj.b == 2
        obj.b = 3

    assert obj.a == 1
    assert obj.b == 3


def test_save_raises_errors():
    from xotl.tools.future.types import SimpleNamespace as new

    getter = lambda o: lambda a: getattr(o, a)
    obj = new(a=1, b=2)
    with pytest.raises(AttributeError):
        with save_attributes(obj, "c", getter=getter):
            pass

    with save_attributes(obj, "x"):
        pass

    assert obj.x is None

    obj = object()
    with pytest.raises(AttributeError):
        with save_attributes(obj, "x"):
            pass


def test_import_object():
    from xotl.tools.objects import import_object

    assert import_object("xotl.tools.objects.import_object") is import_object
    assert import_object("xotl.tools.objects:import_object") is import_object


def test_delegator():
    from xotl.tools.objects import delegator

    class Bar:
        x = object()

    class Foo(delegator("egg", {"x1": "x", "x2": "spam"})):
        def __init__(self):
            self.egg = Bar()

    foo = Foo()
    assert foo.x1 is foo.egg.x

    with pytest.raises(AttributeError):
        foo.x2


def test_final_subclasses():
    from xotl.tools.objects import get_final_subclasses

    class Base:
        pass

    class Subclass(Base):
        pass

    class Final(Subclass):
        pass

    class SubSub(Subclass):
        pass

    class Final2(SubSub):
        pass

    assert set(get_final_subclasses(Base)) == {Final, Final2}
    assert set(get_final_subclasses(Final, include_this=False)) == set([])


def test_FinalSubclassEnumeration():
    from xotl.tools.objects import FinalSubclassEnumeration

    class Base:
        pass

    enum = FinalSubclassEnumeration(Base)
    enum2 = FinalSubclassEnumeration(Base, dynamic=False)

    assert not enum.__members__
    assert not enum2.__members__

    class Subclass(Base):
        pass

    class Final(Subclass):
        pass

    assert enum.Final is Final
    assert not enum2.__members__

    enum2.invalidate_cache()
    assert enum2.Final is Final



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Collection of disparate utilities.

``xotl.tools`` is essentially an extension to the Python's standard library,
it does not make up a full framework, but it's very useful to be used from a
diversity of scenarios.

"""

from ._version import get_versions  # type: ignore

__version__ = get_versions()["version"]
del get_versions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xotl.tools.keywords import getkwd, setkwd, kwd_getter, kwd_setter, org_kwd


def test_keywords():
    class Foobar:
        pass

    proper = lambda v: isinstance(v, type) and issubclass(v, Foobar)
    obj = Foobar()
    names = {"if", "and", "or", "abc", "xyz"}
    for name in names:
        setkwd(obj, name, type(name.title(), (Foobar,), {}))
    for name, value in vars(obj).items():
        if proper(value):
            assert org_kwd(name).title() == value.__name__
    kwd_setter(obj)("else", 123)
    assert obj.if_ is kwd_getter(obj)("if")
    assert obj.and_ is getkwd(obj, "and")
    assert obj.xyz is getkwd(obj, "xyz")
    assert obj.else_ is getkwd(obj, "else")
    keys = {org_kwd(n) for n, v in vars(obj).items() if proper(v)}
    assert keys == names



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import pickle
from functools import total_ordering

from hypothesis import given, strategies as s
from xotl.tools.infinity import Infinity, InfinityType


@given(s.floats() | s.integers())
def test_comparable_with_numbers(x):
    assert -Infinity < x < Infinity


@given(s.dates() | s.datetimes())
def test_comparable_with_dates(x):
    assert -Infinity < x < Infinity


def test_infinity_hashable():
    hash(Infinity)
    hash(-Infinity)


def test_singleton():
    assert Infinity is InfinityType(+1)
    assert -Infinity is InfinityType(-1)


@given(
    s.sampled_from([Infinity, -Infinity]),
    s.sampled_from([pickle.HIGHEST_PROTOCOL, pickle.DEFAULT_PROTOCOL]),
)
def test_pickable(inf, proto):
    serialized = pickle.dumps(inf, proto)
    assert inf is pickle.loads(serialized)


@total_ordering
class WrappedValue:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        if isinstance(other, WrappedValue):
            return self.value == other.value
        else:
            return self.value == other

    def __le__(self, other):
        if isinstance(other, WrappedValue):
            return self.value <= other.value
        else:
            return self.value <= other


@given((s.floats() | s.integers()).map(lambda x: WrappedValue(x)))
def test_comparable_with_wrapped_numbers(x):
    assert -Infinity < x < Infinity


@given((s.dates() | s.datetimes()).map(lambda x: WrappedValue(x)))
def test_comparable_with_wrapped_dates(x):
    assert -Infinity < x < Infinity



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from xotl.tools.modules import moduleproperty
from xotl.tools.objects import memoized_property


@moduleproperty
def this(self):
    return self


@moduleproperty
def store(self):
    return getattr(self, "_store", None)


@store.setter
def store(self, value):
    setattr(self, "_store", value)


@store.deleter
def store(self):
    delattr(self, "_store")


def prop(self):
    return getattr(self, "_prop", None)


def _prop_set(self, val):
    setattr(self, "_prop", val)


def _prop_del(self):
    delattr(self, "_prop")


prop = moduleproperty(prop, _prop_set, _prop_del)


def otherfunction():
    return 1


def memoized(self):
    return self


memoized = moduleproperty(memoized, base=memoized_property)

try:

    @memoized.setter
    def memoized(self, value):
        pass


except AttributeError:
    pass  # Ok
else:
    raise AssertionError("module-level memoized_property should be read-only")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

r"""Facilities to work with `concrete numbers`_.

   A concrete number is a number associated with the things being counted, in
   contrast to an abstract number which is a number as a single entity.

   -- Wikipedia__


__ `concrete numbers`_
.. _concrete numbers: https://en.wikipedia.org/wiki/Concrete_number


This module allows you to define dimensions (or quantity types):

   >>> from xotl.tools.dim.meta import Dimension, UNIT
   >>> @Dimension.new
   ... class Length:
   ...     metre = UNIT
   ...     kilometre = 1000 * metre
   ...     centimetre = metre/100
   ...     milimetre = milimetres = metre/1000
   ...
   ...     inch = inches = 24.5 * milimetres
   ...     foot = feet = 12 * inches


.. seealso:: Module `~xotl.tools.dim.base`:mod: defines the standard base
   quantities.


Each dimension **must** define a single *canonical unit* for measuring
quantities within the dimension.  Values in the dimension are always expressed
in terms of the canonical units.

In the previous example the dimension Length defined the `metre` for its
canonical unit.  The name of canonical unit defines the `signature
<Signature>`:class: for the quantities in the dimension.

When printed (or ``repr``-ed) `quantities <Quantity>`:class: use the format
``<magnitude>::<signature>``:

   >>> metre = Length.metre
   >>> metre
   1::{<Length.metre>}/{}

Quantities support the standard arithmetical operations of addition,
subtraction, multiplication and division.  In fact, you obtain different
quantities in the dimension by multiplying with the canonical unit:

   >>> metre + metre
   2::{<Length.metre>}/{}

   >>> metre*metre
   1::{<Length.metre>, <Length.metre>}/{}

   >>> km = 1000 * metre

   >>> 5 * km
   5000::{<Length.metre>}/{}

`Dimensional homogeneity`__ imposes restrictions on the allowed operations
between quantities.  Only commensurable quantities (quantities of the same
dimension) can be compared, equated, added, or subtracted.

__ https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity

   >>> @Dimension.new
   >>> class Time:
   ...     second = UNIT

   >>> metre + Time.second  # doctest: +ELLIPSIS
   Traceback (...)
   ...
   OperandTypeError: unsupported operand type(s) for +:...


However, you can take ratios of incommensurable quantities (quantities with
different dimensions), and multiply or divide them.

    >>> metre/Time.second
    >>> 1::{<Length.metre>}/{<Time.second>}


.. warning:: `Decimal numbers <decimal.Decimal>`:py:class: are not supported.

   This module makes not attempt to fix the standing incompatibility between
   floats and `decimal.Decimal`:py:class:\ :

      >>> import decimal
      >>> decimal.Decimal('0') + 0.1  # doctest: +ELLIPSIS
      Traceback (...)
      ...
      TypeError: unsupported operand type(s) for +: 'Decimal' and 'float'


The signature created by `Dimension`:class: for its canonical unit is simply a
string that varies with the name of the dimension and that of the canonical
unit.  This implies that you can *recreate* the same dimension and it will be
interoperable with the former::

   >>> @Dimension.new
   ... class L:
   ...    m = UNIT

   >>> m = L.m  # Save this


   >>> # Recreate the same dimension.
   >>> @Dimension.new
   ... class L:
   ...    m = UNIT

   >>> m == L.m
   True

Both the dimension name and the canonical unit name *must* be the same for
this to work.

.. note:: We advice to define a dimension only once and import it where
   needed.

"""
import functools
import numbers
from typing import Any, Callable, FrozenSet, Sequence, Tuple, Optional, overload

from xotl.tools.objects import classproperty
from xotl.tools.future.types import TEq


#: The unit for any kind of quantity.
UNIT = 1


@functools.total_ordering
class Quantity(numbers.Real):
    """A concrete number of `quantity` (expressed in) `units`.

    .. seealso:: https://en.wikipedia.org/wiki/Concrete_number

    :param quantity: A real number.
    :param units: A `signature <Signature>`:class: for the units the
                  denominate the given quantity.

    You can construct instances by operating with the attributes of a
    dimension.  For instance, this is 5 kilometres:

       >>> from xotl.tools.dim.base import L
       >>> 5 * L.km
       5000::{<Length.metre>}/{}

    A concrete number is of the type of its dimension:

       >>> isinstance(5 * L.km, L)
       True

    """

    __slots__ = ("magnitude", "signature")

    def __init__(self, quantity, units):
        if not isinstance(quantity, BareReal):
            raise TypeError("Quantities must be real numbers")
        self.magnitude = quantity
        self.signature = units

    def __str__(self):
        return "{}::{}".format(self.magnitude, self.signature)

    __repr__ = __str__

    def __neg__(self):
        return type(self)(-self.magnitude, self.signature)

    def __pos__(self):
        return type(self)(self.magnitude, self.signature)

    def __add__(self, other):
        if isinstance(other, Quantity) and self.signature == other.signature:
            return type(self)(self.magnitude + other.magnitude, self.signature)
        else:
            # What is the meaning of "10km + 1"?
            raise OperandTypeError("+", self, other)

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Quantity) and self.signature == other.signature:
            return type(self)(self.magnitude - other.magnitude, self.signature)
        else:
            # What is the meaning of "10km - 1"?
            raise OperandTypeError("-", self, other)

    def __mul__(self, other):
        if isinstance(other, BareReal):
            other = type(self)(other, Signature())
        if isinstance(other, Quantity):
            return downgrade_to_scalar(
                type(self)(
                    self.magnitude * other.magnitude, self.signature * other.signature
                )
            )
        else:
            raise OperandTypeError("*", self, other)

    __rmul__ = __mul__

    def __pow__(self, exp):
        if isinstance(exp, numbers.Integral) and exp != 0:
            if exp < 0:
                return 1 / (self ** (-exp))
            else:
                return type(self)(self.magnitude ** exp, self.signature ** exp)
        else:
            raise OperandTypeError("**", self, exp)

    def __div__(self, other):
        if isinstance(other, BareReal):
            other = type(self)(other, Signature())
        if isinstance(other, Quantity):
            return downgrade_to_scalar(
                type(self)(
                    self.magnitude / other.magnitude, self.signature / other.signature
                )
            )
        else:
            raise OperandTypeError("/", self, other)

    __truediv__ = __div__

    def __floordiv__(self, other):
        if isinstance(other, BareReal):
            other = type(self)(other, Signature())
        if isinstance(other, Quantity):
            return downgrade_to_scalar(
                type(self)(
                    self.magnitude // other.magnitude, self.signature / other.signature
                )
            )
        else:
            raise OperandTypeError("//", self, other)

    def __rdiv__(self, other):
        if isinstance(other, BareReal):
            other = type(self)(other, Signature())
            return downgrade_to_scalar(
                type(self)(
                    other.magnitude / self.magnitude, other.signature / self.signature
                )
            )
        else:
            raise OperandTypeError("/", other, self)

    __rtruediv__ = __rdiv__

    def __rfloordiv__(self, other):
        if isinstance(other, BareReal):
            other = type(self)(other, Signature())
            return downgrade_to_scalar(
                type(self)(
                    other.magnitude // self.magnitude, other.signature / self.signature
                )
            )
        else:
            raise OperandTypeError("//", other, self)

    def __hash__(self):
        return hash((self.magnitude, self.signature))

    def __eq__(self, other):
        if isinstance(other, BareReal) and self.signature == SCALAR:
            return self.magnitude == other
        elif isinstance(other, Quantity) and self.signature == other.signature:
            return self.magnitude == other.magnitude
        else:
            raise TypeError("incomparable quantities: %r and %r" % (self, other))

    def __lt__(self, other):
        if isinstance(other, Quantity) and self.signature == other.signature:
            return self.magnitude < other.magnitude
        else:
            raise TypeError("incomparable quantities: %r and %r" % (self, other))

    def __le__(self, other):
        if isinstance(other, Quantity) and self.signature == other.signature:
            return self.magnitude <= other.magnitude
        else:
            raise TypeError("incomparable quantities: %r and %r" % (self, other))

    # The following make Quantity more compatible with numbers.Real.  In all
    # cases, taking a Quantity for a float takes the magnitude expressed in
    # the canonical unit.

    def __float__(self):
        return float(self.magnitude)

    def __trunc__(self):
        return self.magnitude.__trunc__()

    def __abs__(self):
        return abs(self.magnitude)

    def __round__(self):
        return round(self.magnitude)

    def __ceil__(self):
        import math

        return math.ceil(self.magnitude)

    def __floor__(self):
        import math

        return math.floor(self.magnitude)

    def __mod__(self, other):
        if isinstance(other, BareReal):
            return type(self)(self.magnitude % other, self.signature)
        else:
            raise OperandTypeError("%", self, other)

    def __rmod__(self, other):
        # This is a rare operation.  Imagine: 5 % 2m to be 1/m....  But if I
        # can do 5/2m and that is 2.5/m, then % should be allowed.
        if isinstance(other, BareReal):
            return type(self)(other % self.magnitude, 1 / self.signature)
        else:
            raise OperandTypeError("%", self, other)

    def __rpow__(self, other):
        raise OperandTypeError("**", other, self)


class Dimension(type):
    """A type for `quantities`_.

    This is a metaclass for dimensions.  Every instance (class) will
    automatically have the following attributes:

    .. attribute:: _unitname_

       The name of canonical unit in the dimension.  Notice that `aliases
       <new>`:meth: are created after the defined canonical unit.  This is the
       name of the attribute provided in the class definition of the dimension
       with value equal to `UNIT`:const:.

    .. attribute:: _unit_

       The canonical `quantity <Quantity>`:class:.  This is the quantity 1
       (`UNIT`:const:) expressed in terms of the canonical unit.

    .. attribute:: _signature_

       The canonical `signature <Signature>`:class: of the quantities.

    It's always true that ``Quantity(UNIT, self._signature_) == self._unit_``.

    .. _quantities: https://en.wikipedia.org/wiki/Concrete_numbers

    The provided dimension `~xotl.tools.dim.base.Length`:class: has the
    canonical quantity `1 metre`::

      >>> Length.metre
      1::{<Length.metre>}/{}

      >>> Length._unit_ == Length.metre == Quantity(1, Length._signature_)
      True

    You may subclass `Dimension`:class: and customize the attribute
    ``Quantity``.  The value of the attribute **should** be a subclass of
    `~xotl.tools.dim.meta.Quantity`:class: or a callable that returns
    instances of Quantity.  You may provide it by fully-qualified name as
    supported by `~xotl.tools.objects.import_object`:func:.

    .. versionchanged:: 2.1.0 Added class-attribute Quantity.

    """

    #: Customizable Quantity factory.  It must be a callable that takes a
    #: number and a signature, and returns a Quantity-like object.
    Quantity = Quantity

    @classproperty
    def _Quantity(self):
        from xotl.tools.objects import import_object

        return import_object(self.Quantity)

    def __new__(cls, name, bases, attrs):
        wrappedattrs = {}
        Base = next((base for base in bases if isinstance(base, cls)), None)
        if Base is not None:
            unit = Base._unitname_
            signature = Base._signature_
        else:
            unit = None
            signature = Signature()
        for attr, val in attrs.items():
            if isinstance(val, BareReal):
                if val == UNIT and unit is not None:
                    raise TypeError("quantity with multiple units")
                if unit is None and val == UNIT:
                    unit = attr
                    assert not signature.top
                    signature.top = ("<{}.{}>".format(name, unit),)
                    # WARNING: In order to make a single looping structure we
                    # intentionally break the signature immutability
                wrappedattrs[attr] = cls._Quantity(val, signature)
            else:
                if unit is None and isinstance(val, Quantity):
                    # This is the case when I need to create the quantity from
                    # operations.  It's is not a public API.  We also break
                    # the signature immutability here.
                    if val.magnitude == UNIT:
                        unit = attr
                    assert not signature.top and not signature.bottom
                    signature.top = val.signature.top
                    signature.bottom = val.signature.bottom
                wrappedattrs[attr] = val
        if unit is None:
            raise TypeError("dimension without a unit")
        self = super().__new__(cls, name, bases, wrappedattrs)
        self._unitname_ = unit
        self._unit_ = getattr(self, unit)
        self._signature_ = signature
        return self

    @overload
    @classmethod
    def new(cls, source: Any) -> "Dimension":
        ...

    @overload
    @classmethod
    def new(
        cls,
        *,
        unit_alias: str = None,
        unit_aliases: Sequence[str] = None,
        Quantity: type = None
    ) -> Callable[[Any], "Dimension"]:
        ...

    @classmethod
    def new(cls, *source, **kwargs):
        """Define a new dimension.

        This is a wrapped decorator.  The actual possible signatures are:

           - ``new(unit_alias=None, unit_aliases=None, Quantity=None)(source)``

           - ``new(source)``

        This allows to use this method as decorator with or without arguments.

        :param source: A class with at least the canonical unit definition.
                       Other unit definitions will be automatically converted.

        :keyword unit_alias: An alias for the canonical unit.  You cannot use
                             a `source` with several canonical units.  This is
                             a simple way to introduce a single alias.

        :keyword unit_aliases: A sequence with the name of other aliases for
                               the canonical unit.

        :keyword Quantity: A replacement for the default Quantity of the
                 dimension.  It defaults to None; which then looks for the
                 Quantity attribute the class definition.  This allows to
                 provide a custom Quantity without having to subclass
                 Dimension.

        Example:

           >>> @Dimension.new(unit_alias='man')
           ... class Workforce:
           ...    men = UNIT

           >>> Workforce.men == Workforce.man == Workforce._unit_
           True

        The resulting class will be an instance of `Dimension`:class::

           >>> isinstance(Workforce, Dimension)
           True

        The original class is totally missed:

           >>> Workforce.mro()
           [...Workforce, object]

        To complete the example, let's introduce the dimension Effort that
        expresses the usual amount of men-power and time needed to complete
        some task.  However `~xotl.tools.dim.base.Time`:class: has the second
        as it canonical unit, but the standard for Effort is men-hour:

           >>> class Effort(Workforce * Time):
           ...    # Since the canonical unit of a composed quantity type is
           ...    # built from the canonical units of the operands, but the
           ...    # true "canonical type" of effort is usually men-hour we
           ...    # re-introduce it.
           ...    men_hour = 60

        This does not mean that ``Effort._unit_ == Effort.men_hour``.  The
        canonical unit would be ``Effort.men_second``.

        .. versionchanged:: 2.1.0 Added keyword parameter Quantity.

        """
        from xotl.tools.objects import copy_class
        from xotl.tools.decorator.meta import decorator

        @decorator
        def _new(source, unit_alias=None, unit_aliases=None, Quantity=None):
            if Quantity is not None:

                class meta(cls):
                    pass

                # Putting this inside the class fail, because 'Quantity =
                # Quantity' cannot be disambiguated with local and non-local
                # name.
                meta.Quantity = Quantity
            else:
                meta = cls
            res = copy_class(source, meta=meta)
            if unit_alias:
                setattr(res, unit_alias, res._unit_)
            if unit_aliases:
                for alias in unit_aliases:
                    setattr(res, alias, res._unit_)
            return res

        if source and kwargs or len(source) > 1:
            raise TypeError("Invalid signature")
        return _new(*source, **kwargs)

    def __instancecheck__(self, instance):
        if isinstance(instance, Quantity):
            return instance.signature == self._signature_
        else:
            return False

    def __mul__(self, other):
        if isinstance(other, Dimension):
            name = TIMES(self.__name__, other.__name__)
            if self == other:
                unit = SQUARED(self._unitname_)
                quant = self._Quantity(UNIT, self._signature_ ** 2)
            else:
                unit = TIMES(self._unitname_, other._unitname_)
                quant = self._Quantity(UNIT, self._signature_ * other._signature_)
            klass = type(self)
            return klass(name, (object,), {unit: quant})
        else:
            raise OperandTypeError("*", self, other)

    def __pow__(self, exp):
        if isinstance(exp, numbers.Integral):
            if exp == 0:
                return Scalar
            elif exp == 1:
                return self
            elif exp == 2:
                return self * self
            elif exp < 0:
                return 1 / (self ** -exp)
            else:
                assert exp > 0
                name = POWER(self.__name__, exp)
                unit = POWER(self._unitname_, exp)
                quant = self._Quantity(UNIT, self._signature_ ** exp)
                klass = type(self)
                return klass(name, (object,), {unit: quant})
        else:
            raise OperandTypeError("**", self, exp)

    def __div__(self, other):
        if isinstance(other, Dimension):
            if self == other:
                return Scalar
            else:
                name = PER(self.__name__, other.__name__)
                unit = PER(self._unitname_, other._unitname_)
                quant = self._Quantity(UNIT, self._signature_ / other._signature_)
                klass = type(self)
                return klass(name, (object,), {unit: quant})
        else:
            raise OperandTypeError("/", self, other)

    __truediv__ = __floordiv__ = __div__

    def __rdiv__(self, numerator):
        assert not isinstance(numerator, Dimension)
        if numerator == 1:
            name = PER("unit", self.__name__)
            unit = PER("unit", self._unitname_)
            quant = self._Quantity(UNIT, 1 / self._signature_)
            klass = type(self)
            return klass(name, (object,), {unit: quant})
        else:
            raise OperandTypeError("/", numerator, self)

    __rtruediv__ = __rfloordiv__ = __rdiv__

    def __eq__(self, other):
        if isinstance(other, Dimension):
            return self._signature_ == other._signature_
        else:
            raise TypeError(
                "incomparable types '%s' and '%s'"
                % (type(self).__name__, type(other).__name__)
            )


class Signature:
    """The layout of the kinds that compose a quantity.

    The layout is given by a pair non-ordered collections (repetition
    is allowed): the numerator (we call it top within the signature)
    and the denominator (bottom).

    We represent a signature as ``{top elements}/{bottom elements}``.

    You may regard a signature as an abstract 'syntactical' part of a
    quantity.  For Length, the ``{metre}/{}`` is the signature of such a
    quantity.

    The number "10" is not tied to any particular kind of quantity.  Bare
    numbers have no kind and the bear the signature ``{}/{}``.

    The items of top and bottom are required to be comparable for equality
    (``==``).

    You can multiply and divide signatures and simplification happens
    automatically.

    You *should* regard signatures as immutable values.  In fact, this is kind
    of an internal, but interesting, concept of this module.

    Examples::

      >>> distance = Signature('m')
      >>> distance
      {m}/{}

      >>> time = Signature('s')

      >>> freq = 1/time
      >>> freq
      {}/{s}

      >>> speed = distance/time
      >>> speed
      {m}/{s}

      >>> acceleration = speed/time
      >>> acceleration
      {m}/{s, s}

    You may compare signatures for equality.

      >>> acceleration == distance/(time*Signature('s'))
      True

      >>> speed == distance * freq
      True

    Signature don't support neither addition nor subtraction::

      >>> distance + distance  # doctest: +ELLIPSIS
      Traceback (...)
      ...
      TypeError: unsupported operand type(s) for +: 'Signature' and 'Signature'

    """

    __slots__ = ("top", "bottom", "_tps", "_bts")

    def __init__(self, top: Sequence[TEq] = None, bottom: Sequence[TEq] = None):
        self._tps: Optional[FrozenSet] = None
        self._bts: Optional[FrozenSet] = None
        self.top, self.bottom = self.simplify(top, bottom)

    @property
    def _topset(self) -> FrozenSet[Tuple[TEq, int]]:
        res = self._tps
        if res is None:
            from collections import Counter

            counts = Counter(self.top)  # type: ignore
            self._tps = res = frozenset(counts.items())
        return res

    @property
    def _bottomset(self) -> FrozenSet[Tuple[TEq, int]]:
        res = self._bts
        if res is None:
            from collections import Counter

            counts = Counter(self.bottom)  # type: ignore
            self._bts = res = frozenset(counts.items())
        return res

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._topset == other._topset and self._bottomset == other._bottomset
        else:
            return False

    def __hash__(self):
        return hash(self._topset) + hash(self._bottomset)

    def __lt__(self, other):
        raise TypeError("signatures are not orderable")

    __gt__ = __ge__ = __le__ = __lt__

    def __mul__(self, other):
        cls = type(self)
        if other == UNIT:
            return self
        elif isinstance(other, cls):
            return cls(self.top + other.top, self.bottom + other.bottom)
        else:
            raise TypeError

    __rmul__ = __mul__

    def __div__(self, other):
        cls = type(self)
        if other == UNIT:
            return self
        elif isinstance(other, cls):
            return cls(self.top + other.bottom, self.bottom + other.top)
        else:
            raise TypeError

    __truediv__ = __floordiv__ = __div__

    def __rdiv__(self, numerator):
        if numerator == UNIT:
            cls = type(self)
            return cls(self.bottom, self.top)
        else:
            raise TypeError

    __rtruediv__ = __rfloordiv__ = __rdiv__

    def __pow__(self, exp):
        if isinstance(exp, numbers.Integral):
            if exp == 0:
                return Signature()
            elif exp > 0:
                return Signature(self.top * exp, self.bottom * exp)
            else:
                return Signature(self.bottom * -exp, self.top * -exp)
        else:
            raise TypeError

    @staticmethod
    def simplify(top, bottom):
        """Removes equal items from top and bottom in a one-to-one
        correspondence.

        Signatures are simplified on initialization::

           >>> Signature('abcxa', 'bxay')
           {c, a}/{y}

        This function takes top and bottom and returns simplified
        tuples for top and bottom.

        """
        top = [] if top is None else list(top)
        bottom = [] if bottom is None else list(bottom)
        i = 0
        while i < len(top):
            j = 0
            while j < len(bottom) and bottom[j] != top[i]:
                j += 1
            if j < len(bottom):
                assert bottom[j] == top[i]
                del bottom[j]
                del top[i]
            else:
                i += 1
        return tuple(top), tuple(bottom)

    def __str__(self):
        wrap = lambda s: "{{{0}}}".format(s)
        top = wrap(", ".join(str(t) for t in self.top))
        bottom = wrap(", ".join(str(b) for b in self.bottom))
        return "{top}/{bottom}".format(top=top, bottom=bottom)

    __repr__ = __str__


class _BareRealType(type):
    def __instancecheck__(self, i):
        return isinstance(i, numbers.Real) and not isinstance(i, Quantity)


class BareReal(metaclass=_BareRealType):
    """Any real that is not a Quantity instance."""


SCALAR = Signature()


@Dimension.new
class Scalar:
    """A quantity whose signature is always *empty*.

    Most of the time you should not deal with this quantity.  Any normal
    operation that results in a scalar gets reduced to Python's type:

        >>> from xotl.tools.dim.base import L
        >>> L.m/L.m
        1.0

    This type makes the operations on `dimensions <Dimension>`:class: closed
    under multiplication:

        >>> Scalar * L == L == L * Scalar
        True

    """

    unit = Quantity(UNIT, SCALAR)


TIMES = lambda a, b: "{}_{}".format(a, b)
PER = lambda a, b: "{}_per_{}".format(a, b)
SQUARED = lambda a: "{}_squared".format(a)


def POWER(a, e):
    return "{}_pow_{}".format(a, e)


class OperandTypeError(TypeError):
    def __init__(self, operand, val1, val2):
        if isinstance(val1, Quantity):
            t1 = val1.signature
        else:
            t1 = type(val1).__name__
        if isinstance(val2, Quantity):
            t2 = val2.signature
        else:
            t2 = type(val2).__name__
        super().__init__(
            "unsupported operand type(s) for %s: '%s' and '%s'" % (operand, t1, t2)
        )


def downgrade_to_scalar(quantity):
    """Downgrade a concrete number to a bare number if possible.

    .. note:: This is not an API of this module.

    """
    if quantity.signature == SCALAR:
        return quantity.magnitude
    else:
        return quantity



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Records definitions.

A record allows to describe plain external data and a simplified model to
*read* it.  The main use of records is to represent data that is read from a
`CSV  file <305>`:pep:.

See the `record`:class: class to find out how to use it.

"""


from xotl.tools.symbols import Unset
from xotl.tools.future.functools import lru_cache


@lru_cache()
def field_descriptor(field_name):
    """Returns a read-only descriptor for `field_name`."""

    class descriptor:
        def __get__(self, instance, owner):
            if instance:
                return owner.get_field(
                    instance._raw_data, owner._rec_fields[field_name]
                )
            else:
                return self

    return descriptor


class _record_type(type):
    @staticmethod
    def _is_rec_definition(attr, val=Unset):
        result = not attr.startswith("_") and attr.upper() == attr
        if val is not Unset:
            from numbers import Integral

            result = result and isinstance(val, (Integral, str))
        return result

    @staticmethod
    def is_reader(attr, func, fields=None):
        from xotl.tools.future.types import FunctionType

        attr = attr.lower()
        good_name = attr.startswith("_") and attr.endswith("_reader")
        good_type = isinstance(func, (FunctionType, staticmethod))
        return good_name and good_type

    def __new__(cls, name, bases, attrs):
        def static(f):
            return f if isinstance(f, staticmethod) else staticmethod(f)

        cls_fields = {
            attr: val
            for attr, val in attrs.items()
            if cls._is_rec_definition(attr, val)
        }
        descriptors = {attr.lower(): field_descriptor(attr)() for attr in cls_fields}
        readers = {
            attr.lower(): static(func)
            for attr, func in attrs.items()
            if cls.is_reader(attr, func)
        }
        new_attrs = dict(attrs, **descriptors)
        new_attrs.update(readers)
        result = super().__new__(cls, name, bases, new_attrs)
        # Make a copy, or else the super-class attribute gets contaminated
        fields = dict(getattr(result, "_rec_fields", {}))
        index = dict(getattr(result, "_rec_index", {}))
        fields.update(cls_fields)
        if len(fields) != len({val for val in fields.values()}):
            msg = 'Duplicated field index definition in class "%s"' % name
            import logging

            logger = logging.getLogger(__name__)
            logger.error(msg)
            logger.debug(fields)
            raise TypeError(msg)
        result._rec_fields = fields
        index.update({val: attr for attr, val in cls_fields.items()})
        result._rec_index = index
        return result

    def get_field(self, raw_data, field):
        from xotl.tools.symbols import Undefined

        field_name = self._rec_index[field]
        try:
            value = raw_data[field]
        except (IndexError, KeyError):
            value = Undefined
        reader_name = "_%s_reader" % field_name.lower()
        reader = getattr(self, reader_name, None)
        if reader:
            return reader(value)
        else:
            return value


class record(metaclass=_record_type):
    """Base record class.

    Records allow to represent a sequence or mapping of values extracted from
    external sources into a dict-like Python value.

    The first use-case for this abstraction is importing data from a `CSV file
    <305>`:pep:.  You could represent each line as an instance of a properly
    defined record.

    An instance of a record would represent a single `line` (or row) from the
    external data source.

    Records are expected to declare `fields`.  Each field must be a
    CAPITALIZED valid identifier like::

        >>> class INVOICE(record):
        ...     ID = 0
        ...     REFERENCE = 1

    Fields must be integers or plain strings.  Fields must not begin with an
    underscore ("_").  External data lines are required to support indexes of
    those types.

    You could use either the classmethod `get_field`:func: to get the value of
    field in a single line (data as provided by the external source)::

        >>> line = (1, 'AA20X138874Z012')
        >>> INVOICE.get_field(line, INVOICE.REFERENCE)
        'AA20X138874Z012'

    You may also have an instance::

        >>> invoice = INVOICE(line)
        >>> invoice.reference
        'AA20X138874Z012'

    .. note:: Instances attributes are renamed to lowercase.  So you **must**
       not create any other attribute that has the same name as a field in
       lowercase, or else it will be overwritten.

    You could define `readers` for any field.  For instance if you have a
    "CREATED_DATETIME" field you may create a "_created_datetime_reader"
    function that will be used to parse the raw value of the instance into an
    expected type.  See the `included readers builders below
    <included-readers>`:ref:.

    Readers are always cast as `staticmethods`, whether or not you have
    explicitly stated that fact::

        >>> from dateutil import parser
        >>> class BETTER_INVOICE(INVOICE):
        ...     CREATED_TIME = 2
        ...     _created_time_reader = lambda val: parser.parse(val)

        >>> line = (1, 'AA20X138874Z012', '2014-02-17T17:29:21.965053')
        >>> BETTER_INVOICE.get_field(line, BETTER_INVOICE.CREATED_TIME)
        datetime.datetime(2014, 2, 17, 17, 29, 21, 965053)

    .. warning:: Creating readers for fields defined in super classes is not
       directly supported.  To do so, you **must** declare the reader as a
       staticmethod yourself.

    .. note:: Currently there's no concept of relationship between rows in
       this model.  We are evaluating whether by placing a some sort of
       context into the `kwargs` argument would be possible to write readers
       that fetch other instances.

    """

    def __init__(self, raw_data):
        self._raw_data = raw_data

    def __repr__(self):
        cls = type(self)
        return "%s(%r)" % (cls.__name__, self._raw_data)

    def __getitem__(self, field_index):
        return type(self).get_field(self._raw_data, field_index)


def isnull(val):
    """Return True if `val` is null.

    Null values are None, the empty string and any False instance of
    `xotl.tools.symbols.boolean`:class:.

    Notice that 0, the empty list and other false values in Python are not
    considered null.  This allows that the CSV null (the empty string) is
    correctly treated while other sources that provide numbers (and 0 is a
    valid number) are not misinterpreted as null.

    """
    from xotl.tools.symbols import boolean

    return val in (None, "") or (isinstance(val, boolean) and not val)


# Standard readers
def check_nullable(val, nullable):
    """Check the restriction of nullable.

    Return True if the val is non-null.  If nullable is True and the val is
    null returns False.  If `nullable` is False and `val` is null, raise a
    ValueError.

    Test for null is done with function `isnull`:func:.

    """
    null = isnull(val)
    if not null or nullable:
        return not null
    else:
        raise ValueError("NULL value was not expected here")


@lru_cache()
def datetime_reader(format, nullable=False, default=None, strict=True):
    """Returns a datetime reader.

    :param format: The format the datetime is expected to be in the external
       data.  This is passed to `datetime.datetime.strptime`:func:.

    :param strict: Whether to be strict about datetime format.

    The reader works first by passing the value to strict
    `datetime.datetime.strptime`:func: function.  If that fails with a
    ValueError and strict is True the reader fails entirely.

    If strict is False, the worker applies different rules.  First if the
    `dateutil` package is installed its parser module is tried.  If `dateutil`
    is not available and nullable is True, return None; if nullable is False
    and default is not null (as in `isnull`:func:), return `default`,
    otherwise raise a ValueError.

    .. versionadded: 1.6.7  Add the `strict` argument.

    .. versionchanged: 1.6.7.1  Keep the meaning of null when testing for
       `default` if strict is False and dateutil is not available.

    """
    try:
        from dateutil.parser import parse
    except ImportError:
        parse = None

    def reader(val):
        if check_nullable(val, nullable):
            from datetime import datetime

            try:
                return datetime.strptime(val, format)
            except ValueError:
                if strict:
                    raise
                elif parse:
                    return parse(val)
                else:
                    if nullable:
                        return None
                    elif not isnull(default):
                        return default
                    else:
                        raise ValueError
        else:
            return default

    return reader


@lru_cache()
def date_reader(format, nullable=False, default=None, strict=True):
    """Return a date reader.

    This is similar to `datetime_reader`:func: but instead of returning a
    `datetime.datetime`:class: it returns a `datetime.date`.

    Actually this function delegates to `datetime_reader`:func: most of its
    functionality.

    .. versionadded: 1.6.8

    """
    reader = datetime_reader(format, nullable=nullable, default=default, strict=strict)

    def res(val):
        result = reader(val)
        if not isnull(result) and result is not default:
            return result.date()
        else:
            return result

    return res


@lru_cache()
def boolean_reader(true=("1",), nullable=False, default=None):
    """Returns a boolean reader.

    :param true: A collection of raw values considered to be True.  Only the
                 values in this collection will be considered True values.

    """

    def reader(val):
        if check_nullable(val, nullable):
            return val in true
        else:
            return default

    return reader


@lru_cache()
def integer_reader(nullable=False, default=None):
    """Returns an integer reader."""

    def reader(val):
        if check_nullable(val, nullable):
            return int(val)
        else:
            return default

    return reader


@lru_cache()
def decimal_reader(nullable=False, default=None):
    """Returns a Decimal reader."""

    def reader(val):
        if check_nullable(val, nullable):
            from decimal import Decimal

            return Decimal(val)
        else:
            return default

    return reader


@lru_cache()
def float_reader(nullable=False, default=None):
    """Returns a float reader."""

    def reader(val):
        if check_nullable(val, nullable):
            return float(val)
        else:
            return default

    return reader


del lru_cache



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extensions to os.path

Functions inside this module must not have side-effects on the
file-system. This module re-exports (without change) several functions from the
`os.path`:mod: standard module.

"""

import sys
from os.path import abspath, expanduser, dirname, sep, normpath, join as _orig_join

from xotl.tools.future.functools import power as pow_

__all__ = (
    "abspath",
    "expanduser",
    "dirname",
    "sep",
    "normpath",
    "rtrim",
    "fix_encoding",
    "join",
    "normalize_path",
    "shorten_module_filename",
    "shorten_user",
)


# TODO: import all in "from os.path import *"


def rtrim(path, n=1):
    """Trims the last `n` components of the pathname `path`.

    This basically applies `n` times the function `os.path.dirname` to `path`.

    `path` is normalized before proceeding (but not tested to exists).

    .. versionchanged:: 1.5.5 `n` defaults to 1.  In this case rtrim is
                        identical to `os.path.dirname`:func:.

    Example::

      >>> rtrim('/tmp/a/b/c/d', 3)
      '/tmp/a'

      # It does not matter if `/` is at the end
      >>> rtrim('/tmp/a/b/c/d/', 3)
      '/tmp/a'

    """
    return pow_(dirname, n)(normalize_path(path))


def fix_encoding(name, encoding=None):
    """Fix encoding of a file system resource name.

    `encoding` is ignored if `name` is already a `str`.

    """
    if not isinstance(name, str):
        if not encoding:
            from xotl.tools.future.codecs import force_encoding

            encoding = force_encoding(sys.getfilesystemencoding())
        fixer = name.decode if isinstance(name, bytes) else name.encode
        return fixer(encoding)
    else:
        return name


def join(base, *extras):
    """Join two or more pathname components, inserting '/' as needed.

    If any component is an absolute path, all previous path components
    will be discarded.

    Normalize path (after join parts), eliminating double slashes, etc.

    """
    try:
        path = _orig_join(base, *extras)
    except Exception:  # TODO: @med which exceptions expected?
        base = fix_encoding(base)
        extras = [fix_encoding(extra) for extra in extras]
        path = _orig_join(base, *extras)
    return normpath(path)


def normalize_path(base, *extras):
    """Normalize path by:

    - expanding '~' and '~user' constructions.
    - eliminating double slashes
    - converting to absolute.

    """
    # FIXME: [med] Redundant "path" in name "xotl.tools.fs.path.normalize_path"
    try:
        path = _orig_join(base, *extras)
    except Exception:  # TODO: @med which exceptions expected?
        path = join(base, *extras)
    return abspath(expanduser(path))


def shorten_module_filename(filename):
    """A filename, normally a module o package name, is shortened looking his
    head in all python path.

    """
    path = sys.path[:]
    path.sort(lambda x, y: len(y) - len(x))
    for item in path:
        if item and filename.startswith(item):
            filename = filename[len(item) :]
            if filename.startswith(sep):
                filename = filename[len(sep) :]
    for item in ("__init__.py", "__init__.pyc"):
        if filename.endswith(item):
            filename = filename[: -len(item)]
            if filename.endswith(sep):
                filename = filename[: -len(sep)]
    return shorten_user(filename)


def shorten_user(filename):
    """A filename is shortened looking for the (expantion) $HOME in his head
    and replacing it by '~'.

    """
    home = expanduser("~")
    if filename.startswith(home):
        filename = _orig_join("~", filename[len(home) :])
    return filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extensions the `subprocess` module in the standard library."""


from subprocess import *  # noqa
from subprocess import __all__, Popen, PIPE  # noqa

__all__ = list(__all__) + ["call_and_check_output"]


def call_and_check_output(*popenargs, **kwargs):
    """Combines `call` and `check_output`. Returns a tuple ``(returncode,
    output, err_output)``.

    """
    if "stdout" in kwargs:
        raise ValueError("stdout argument not allowed, it will be overridden.")
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, err = process.communicate()
    retcode = process.poll()
    return (retcode, output, err)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import os
import unittest
import tempfile
import shutil

# Makes sure our package is always importable
# sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
#                                              '..', '..')))


class TestFs(unittest.TestCase):
    def setUp(self):
        # Makes all names predictable
        pjoin = os.path.join
        self.previous_dir = os.getcwd()
        self.base = base = tempfile.mkdtemp(prefix="xotl-tools-tests-")
        os.makedirs(pjoin(base, "A", "B", "C"))
        os.makedirs(pjoin(base, "A", "D", "E"))
        os.makedirs(pjoin(base, "A", "F"))
        self.files = files = []

        wexpected = self.walk_up_expected = pjoin(self.base, "A")
        sentinel = tempfile.mkstemp(prefix="X", dir=wexpected)
        self.sentinel = os.path.basename(sentinel[-1])
        files.append(sentinel)  # For testing `walk_up`
        files.append(tempfile.mkstemp(prefix="M", dir=pjoin(self.base, "A", "B")))
        files.append(tempfile.mkstemp(prefix="P", dir=pjoin(self.base, "A", "B")))
        wstart = self.walk_up_start = pjoin(self.base, "A", "B", "C")
        files.append(tempfile.mkstemp(prefix="z", dir=wstart))
        files.append(tempfile.mkstemp(suffix="ending", dir=pjoin(self.base, "A", "D")))
        files.append(tempfile.mkstemp(prefix="Z", dir=pjoin(self.base, "A", "F")))

    def test_iter_files_with_regex_pattern(self):
        from xotl.tools.fs import iter_files

        res = list(iter_files(self.base, "(?xi)/Z"))
        self.assertEqual(2, len(res))
        self.assertIn(self.files[-3][-1], res)
        self.assertIn(self.files[-1][-1], res)

    def test_iter_files_with_maxdepth(self):
        from xotl.tools.fs import iter_files

        res = list(iter_files(self.base, "(?xi)/Z", maxdepth=3))
        self.assertEqual(1, len(res))
        self.assertIn(self.files[-1][-1], res)
        res = list(iter_files(self.base, "(?xi)/Z", maxdepth=2))
        self.assertEqual(0, len(res))

    def test_walk_up(self):
        from xotl.tools.fs import walk_up

        expected, start = self.walk_up_expected, self.walk_up_start
        sentinel = self.sentinel
        self.assertEqual(expected, walk_up(start, sentinel))

    def test_ensure_filename(self):
        from xotl.tools.fs import ensure_filename

        filename = os.path.join(self.base, "en", "sure", "filename.txt")
        ensure_filename(filename)
        self.assertTrue(os.path.isfile(filename))

    def tearDown(self):
        shutil.rmtree(self.base)
        os.chdir(self.previous_dir)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#


def test_ppformat_rtype():
    from xotl.tools.future.pprint import ppformat

    o = [list(range(i + 1)) for i in range(10)]
    assert type(ppformat(o)) is str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import sys
import unittest

from xotl.tools.modules import customize, modulemethod


class TestModulesCustomization(unittest.TestCase):
    def setUp(self):
        from . import testbed

        self.testbed = testbed

    def tearDown(self):
        sys.modules[self.testbed.__name__] = self.testbed

    def test_echo(self):
        from . import testbed

        module, created, klass = customize(testbed)
        self.assertEqual(10, module.echo(10))

    def test_module_props(self):
        @property
        def this(mod):
            return mod

        from . import testbed

        attrs = {"this": this}
        module, created, klass = customize(testbed, custom_attrs=attrs)
        self.assertEqual(module, module.this)


class TestModuleDecorators(unittest.TestCase):
    def test_echo_module_level(self):
        import sys

        @modulemethod
        def echo(self, *args):
            return (self, args)

        current_module = sys.modules[__name__]
        self.assertEqual((current_module, (1, 2)), echo(1, 2))

    def test_moduleproperties(self):
        from . import customizetestbed as m

        self.assertIs(m, m.this)
        self.assertIs(None, m.store)
        self.assertIsNone(m.prop)
        m.store = (1, 2)
        m.prop = "prop"
        self.assertEqual((1, 2), m.store)
        self.assertEqual((1, 2), m._store)
        self.assertIs("prop", m.prop)

        with self.assertRaises(AttributeError):
            m.this = 1

        del m.store
        with self.assertRaises(AttributeError):
            m._store == ()
        self.assertIs(None, m.store)

        del m.prop
        with self.assertRaises(AttributeError):
            m._prop == "prop"
        self.assertIsNone(m.prop)

    def test_module_level_memoized_props(self):
        from . import customizetestbed as m
        from xotl.tools.future.inspect import getattr_static

        self.assertNotEquals(getattr_static(m, "memoized"), m)
        self.assertIs(m.memoized, m)
        self.assertIs(getattr_static(m, "memoized"), m)


def test_get_module_path_by_module_object():
    import xotl.tools
    import xotl.tools.future.itertools
    from os.path import join
    from xotl.tools.modules import get_module_path

    top = xotl.tools.__path__[0]
    expected = top
    assert get_module_path(xotl.tools) == expected

    expected = (
        join(top, "future", "itertools.py"),
        join(top, "future", "itertools.pyc"),
        join(top, "future", "itertools.pyo"),
    )
    assert get_module_path(xotl.tools.future.itertools) in expected


def test_get_module_path_by_module_string_abs():
    import xotl.tools
    from os.path import join
    from xotl.tools.modules import get_module_path

    top = xotl.tools.__path__[0]
    expected = top
    assert get_module_path("xotl.tools") == expected
    expected = (
        join(top, "future", "itertools.py"),
        join(top, "future", "itertools.pyc"),
        join(top, "future", "itertools.pyo"),
    )
    assert get_module_path("xotl.tools.future.itertools") in expected


def test_get_module_path_by_module_string_rel():
    import pytest
    from xotl.tools.modules import get_module_path

    with pytest.raises(TypeError):
        assert get_module_path(".iterators")


def test_object_stability():
    from . import testbed
    from .testbed import selfish

    a, b = testbed.selfish()
    c, d = selfish()
    e, f = testbed.selfish()
    assert a == c == e
    assert b == d == f



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extensions to the `json` standard library module.

It just adds the ability to encode/decode datetimes. But you should use the
JSONEncoder yourself.

You may use this module as drop-in replacement to Python's `json`.  Also it
contains definitions to use C library JSON speedups or Python replacements in
case that library is not installed in your system.

"""

# TODO: consider use IoC to extend python json module

from json import *  # noqa
import json as _stdlib  # noqa

from json import __all__  # noqa

__all__ = list(__all__) + ["file_load", "encode_string"]

from json import encoder, decoder  # noqa


class JSONEncoder(_stdlib.JSONEncoder):
    __doc__ = (
        _stdlib.JSONEncoder.__doc__
        + """
    xotl.tools extends this class by supporting the following data-types (see
    `default`:meth: method):

    - `datetime`, `date` and `time` values, which are translated to strings
      using ISO format.

    - `Decimal` values, which are represented as a string representation.

    - Iterables, which are represented as lists.

    """
    )
    DATE_FORMAT = str("%Y-%m-%d")
    TIME_FORMAT = str("%H:%M:%S")
    DT_FORMAT = str("%s %s") % (DATE_FORMAT, TIME_FORMAT)

    def default(self, obj):
        from datetime import datetime, date, time
        from decimal import Decimal
        from collections.abc import Iterable

        if isinstance(obj, datetime):
            return obj.strftime(self.DT_FORMAT)
        elif isinstance(obj, date):
            return obj.strftime(self.DATE_FORMAT)
        elif isinstance(obj, time):
            return obj.strftime(self.TIME_FORMAT)
        elif isinstance(obj, Decimal):
            return str(obj)
        elif isinstance(obj, Iterable):
            return list(iter(obj))
        return super().default(obj)


try:
    # New in version 3.5 of standard module.
    JSONDecodeError  # noqa
except NameError:
    # Previous implementations raise 'ValueError'
    JSONDecodeError = ValueError


def file_load(filename):
    with file(filename, "r") as f:
        return load(f)  # noqa


# --- encode strings ---

from json.encoder import encode_basestring  # noqa

try:
    from _json import encode_basestring_ascii
except ImportError:
    from json.encoder import (
        py_encode_basestring_ascii as encode_basestring_ascii,  # noqa
    )


def encode_string(string, ensure_ascii=True):
    """Return a JSON representation of a Python string.

    :param ensure_ascii: If True, the output is guaranteed to be of type
        `str` with all incoming non-ASCII characters escaped.  If False, the
        output can contain non-ASCII characters.

    """
    encode = encode_basestring_ascii if ensure_ascii else encode_basestring
    return encode(string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""This module simply serves the purposes of the testing weaving modules"""

from xotl.tools.modules import moduleproperty, modulemethod


def echo(what):
    return what


@moduleproperty
def this(self):
    return self


def rien():
    return 1


@modulemethod
def method(self, *args, **kwargs):
    return self, args, kwargs


@modulemethod
def selfish(self):
    return self.selfish, selfish



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Tools for Command-Line Interface (CLI) applications.

CLI is a mean of interaction with a computer program where the user (or
client) issues commands to the program in the form of successive lines of text
(command lines).

Commands can be registered by:

  - sub-classing the `Command`:class:,
  - using `~abc.ABCMeta.register`:meth: ABC mechanism for virtual sub-classes,
  - redefining `~`Command.sub_commands`` class method.

.. versionadded:: 1.4.1

"""

from abc import abstractmethod, ABCMeta, ABC
from xotl.tools.objects import staticproperty
from xotl.tools.cli.tools import command_name, program_name


class CommandMeta(ABCMeta):
    """Meta-class for all commands."""

    def __new__(meta, name, bases, namespace):
        cls = super(CommandMeta, meta).__new__(meta, name, bases, namespace)
        cls.__subcommands_registry__ = set()
        return cls

    def register(cls, subclass):
        """Register a virtual subclass of a Command.

        Returns the sub-command, to allow usage as a class decorator.

        .. note:: Python 3.7 hides internal registry (``_abc_registry``), so
                  a sub-commands registry is implemented.

        """
        cls.__subcommands_registry__.add(subclass)
        res = super(CommandMeta, cls).register(subclass)
        if res is None:
            res = subclass
        return res

    def cli_name(cls):
        """Calculate the command name.

        Standard method uses `~xotl.tools.cli.tools.hyphen_name`.  Redefine it
        to obtain a different behaviour.

        Example::

            >>> class MyCommand(Command):
            ...     pass

            >>> MyCommand.cli_name() == 'my-command'
            True

        """
        from xotl.tools.cli.tools import hyphen_name

        unset = object()
        names = ("command_cli_name", "__command_name__")
        i, res = 0, unset
        while i < len(names) and res is unset:
            name = names[i]
            res = getattr(cls, names[i], unset)
            if res is unset:
                i += 1
            elif not isinstance(res, str):
                msg = "Attribute '{}' must be a string.".format(name)
                raise TypeError(msg)
        if res is unset:
            res = hyphen_name(cls.__name__)
        return res

    def get_setting(cls, name, *default):
        aux = len(default)
        if aux < 2:
            unset = object()
            default = default[0] if aux == 1 else unset
            res = cls.__settings__.get(name, default)
            if res is not unset:
                return res
            else:
                raise KeyError(name)
        else:
            msg = "get_setting() takes at most 3 arguments ({} given)"
            raise TypeError(msg.format(aux + 2))

    def set_setting(cls, name, value):
        cls.__settings__[name] = value  # TODO: Check type

    def set_default_command(cls, cmd=None):
        """Default command is called when no one is specified.

        A command is detected when its name appears as the first command-line
        argument.

        To specify a default command, use this method with the command as a
        string (the command name) or the command class.

        If the command is specified, then the calling class is the selected
        one.

        For example::

            >>> Command.set_default_command('server')  # doctest: +SKIP
            >>> Server.set_default_command()           # doctest: +SKIP
            >>> Command.set_default_command(Server)    # doctest: +SKIP

        """
        if cls is Command:
            if cmd is not None:
                name = cmd if isinstance(cmd, str) else command_name(cmd)
            else:
                # TODO: consider reset to None
                raise ValueError("missing command specification!")
        else:
            if cmd is None:
                name = command_name(cls)
            else:
                raise ValueError("redundant command specification", cls, cmd)
        Command.set_setting("default_command", name)


class Command(ABC, metaclass=CommandMeta):
    """Base for all commands."""

    __settings__ = {
        # 'default_command' : None
    }
    __registry_cache__ = {}

    def __str__(self):
        return command_name(type(self))

    def __repr__(self):
        return "<command: %s>" % command_name(type(self))

    @staticproperty
    def registry():
        """Obtain all registered commands."""
        res = Command.__registry_cache__
        if not res:
            Command._settle_cache(Command)
            assert res.pop(command_name(Command), None) is None
            Command._check_help()
        return res

    @abstractmethod
    def run(self, args=None):
        '''Must return a valid value for "sys.exit"'''
        raise NotImplementedError

    @staticmethod
    def _settle_cache(source, recursed=None):
        """Initialize '__registry_cache__'."""
        from xotl.tools.names import nameof

        if recursed is None:
            recursed = set()
        name = nameof(source, inner=True, full=True)
        if name not in recursed:
            recursed.add(name)
            sub_commands = type.__subclasses__(source)
            virtuals = getattr(source, "__subcommands_registry__", ())
            sub_commands.extend(virtuals)
            cmds = getattr(source, "__commands__", None)
            if cmds:
                from collections.abc import Iterable

                if not isinstance(cmds, Iterable):
                    cmds = cmds()
                sub_commands.extend(cmds)
            if sub_commands:
                for cmd in sub_commands:
                    Command._settle_cache(cmd, recursed=recursed)
            else:  # Only branch commands are OK to execute
                from types import FunctionType as ValidMethodType

                assert isinstance(
                    source.run, ValidMethodType
                ), "Invalid type %r for source %r" % (
                    type(source.run).__name__,
                    source,
                )
                Command.__registry_cache__[command_name(source)] = source
        else:
            raise ValueError('Reused class "%s"!' % name)

    @staticmethod
    def _check_help():
        """Check that correct help command is present."""
        name = HELP_NAME
        hlp = Command.__registry_cache__[name]
        if hlp is not Help and not getattr(hlp, "__overwrite__", False):
            Command.__registry_cache__[name] = Help


class Help(Command):
    """Show all commands.

    Define the class attribute `__order__` to sort commands in special command
    "help".

    Commands could define its help in the first line of a sequence of
    documentations until found:

      - command class,
      - "run" method,
      - definition module.

    This command could not be overwritten unless using the class attribute:

       __overwrite__ = True

    """

    __order__ = -9999

    @classmethod
    def get_arg_parser(cls):
        """This is an example on how to build local argument parser.

        Use class method "get

        """
        # TODO: Use 'add_subparsers' in this logic (see 'backlog.org').
        res = getattr(cls, "_arg_parser")
        if not res:
            from argparse import ArgumentParser

            res = ArgumentParser()
            cls._arg_parser = res
        return res

    def run(self, args=[]):
        print('The most commonly used "%s" commands are:' % program_name())
        cmds = Command.registry
        ordered = [(getattr(cmds[cmd], "__order__", 0), cmd) for cmd in cmds]
        ordered.sort()
        max_len = len(max(ordered, key=lambda x: len(x[1]))[1])
        for _, cmd in ordered:
            cmd_class = cmds[cmd]
            doc = self._strip_doc(cmd_class.__doc__)
            if not doc:
                doc = self._strip_doc(cmd_class.run.__doc__)
            if not doc:
                import sys

                mod_name = cmd_class.__module__
                module = sys.modules.get(mod_name, None)
                if module:
                    doc = self._strip_doc(module.__doc__)
                    doc = '"%s"' % (doc if doc else mod_name)
                else:
                    doc = '"%s"' % mod_name
            head = " " * 3 + cmd + " " * (2 + max_len - len(cmd))
            print(head, doc)

    @staticmethod
    def _strip_doc(doc):
        if doc:
            doc = str("%s" % doc).strip()
            return str(doc.split("\n")[0].strip(""""' \t\n\r"""))
        else:
            return ""


HELP_NAME = command_name(Help)

# TODO: Create "xotl.tools.config" here

del abstractmethod, ABCMeta
del staticproperty
